# -*- coding: utf-8 -*-
#
# Copyright (C) 2024 CERN.
# Copyright (C) 2024 Uni Münster.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Vocabulary results."""

from flask import current_app
from invenio_records_resources.proxies import current_service_registry
from invenio_records_resources.services import RecordServiceConfig
from invenio_records_resources.services.records.results import RecordList
from invenio_search import current_search_client

from ..proxies import current_service


class VocabularyTypeList(RecordList):
    """Ensures that vocabulary type metadata is returned in the proper format."""

    @property
    def total(self):
        """Get total number of hits."""
        return self._results.total

    @property
    def custom_vocabulary_names(self):
        """Checks whether vocabulary is a custom vocabulary."""
        return current_app.config.get("VOCABULARIES_CUSTOM_VOCABULARY_TYPES", [])

    def to_dict(self):
        """Formats result to a dict of hits."""
        config_vocab_types = current_app.config.get(
            "INVENIO_VOCABULARY_TYPE_METADATA", {}
        )

        count_terms_agg = {}
        generic_stats = self._generic_vocabulary_statistics()
        custom_stats = self._custom_vocabulary_statistics()

        for k in generic_stats.keys() | custom_stats.keys():
            count_terms_agg[k] = generic_stats.get(k, 0) + custom_stats.get(k, 0)

        hits = self._results.items

        # Extend database data with configuration & aggregation data.
        results = []
        for db_vocab_type in hits:
            result = {
                "id": db_vocab_type.id,
                "pid_type": db_vocab_type.pid_type,
                "count": count_terms_agg.get(db_vocab_type.id, 0),
                "is_custom_vocabulary": db_vocab_type.id
                in self.custom_vocabulary_names,
            }

            if db_vocab_type.id in config_vocab_types:
                for k, v in config_vocab_types[db_vocab_type.id].items():
                    result[k] = v

            results.append(result)

        for hit in results:
            if self._links_item_tpl:
                hit["links"] = self._links_item_tpl.expand(self._identity, hit)

        res = {
            "hits": {
                "hits": results,
                "total": self.total,
            }
        }

        if self._params:
            if self._links_tpl:
                res["links"] = self._links_tpl.expand(self._identity, self.pagination)

        return res

    def _custom_vocabulary_statistics(self):
        # query database for count of terms in custom vocabularies
        returndict = {}
        for vocab_type in self.custom_vocabulary_names:
            custom_service = current_service_registry.get(vocab_type)
            record_cls = custom_service.config.record_cls
            returndict[vocab_type] = record_cls.model_cls.query.count()

        return returndict

    def _generic_vocabulary_statistics(self):
        # Opensearch query for generic vocabularies
        config: RecordServiceConfig = (
            current_service.config
        )  # TODO: Where to get the config from here? current_service is None
        search_opts = config.search

        search = search_opts.search_cls(
            using=current_search_client,
            index=config.record_cls.index.search_alias,
        )

        search.aggs.bucket("vocabularies", {"terms": {"field": "type.id", "size": 100}})

        search_result = search.execute()
        buckets = search_result.aggs.to_dict()["vocabularies"]["buckets"]

        return {bucket["key"]: bucket["doc_count"] for bucket in buckets}



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2022-2024 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test custom fields."""

import pytest
from invenio_records_resources.records.systemfields import PIDListRelation, PIDRelation
from marshmallow import Schema, ValidationError

from invenio_vocabularies.services.custom_fields import VocabularyCF


@pytest.fixture(scope="module")
def vocabulary_cf():
    return VocabularyCF("test", "test")


@pytest.fixture(scope="module")
def TestSchema(vocabulary_cf):
    return Schema.from_dict({"test": vocabulary_cf.field})


@pytest.fixture(scope="module")
def TestUISchema(vocabulary_cf):
    return Schema.from_dict({"test": vocabulary_cf.ui_field})


def test_relation_cls(vocabulary_cf):
    assert vocabulary_cf.relation_cls == PIDRelation

    multi = VocabularyCF("test", "test", multiple=True)
    assert multi.relation_cls == PIDListRelation


def test_cf_mapping(vocabulary_cf):
    # this will be useful when implementing compatibility with OS
    assert vocabulary_cf.mapping == {
        "type": "object",
        "properties": {
            "@v": {"type": "keyword"},
            "id": {"type": "keyword"},
            "title": {"type": "object", "dynamic": "true"},
        },
    }


def test_field_load(TestSchema):
    vocab = {"test": {"id": "test", "title": {"en": "Test"}}}
    assert TestSchema().load(vocab) == {"test": {"id": "test"}}


def test_field_dump(TestSchema):
    vocab = {"test": {"id": "test", "title": {"en": "Test"}}}
    assert TestSchema().load(vocab) == {"test": {"id": "test"}}


def test_ui_field_dump(app, TestUISchema):
    # app is needed for babel local setup
    # no load test since it's never used for loading
    vocab = {
        "test": {
            "id": "test",
            "title": {"en": "Test"},
            "description": {"en": "Test description"},
            "props": {"key": "value"},
            "icon": "icon.png",
            "tags": ["tag1", "tag2"],
        }
    }

    expected_vocab = {
        "test": {
            "id": "test",
            "title_l10n": "Test",
            "description_l10n": "Test description",
            "props": {"key": "value"},
            "icon": "icon.png",
            "tags": ["tag1", "tag2"],
        }
    }
    assert TestUISchema().dump(vocab) == expected_vocab


def test_options(lang_data_many, identity):
    cf = VocabularyCF("test", "languages")
    expected_options = [
        {
            "id": lang,
            "title_l10n": "English",
            "icon": "file-o",
            "props": {
                "akey": "avalue",
            },
        }
        for lang in ["fr", "tr", "gr", "ger", "es"]
    ]

    assert cf.options(identity) == expected_options


def test_vocab_field_required():
    cf = VocabularyCF("test", "test", field_args={"required": True})
    schema = Schema.from_dict({"test": cf.field})()

    with pytest.raises(ValidationError):
        schema.load({})


def test_vocab_ui_field_required():
    cf = VocabularyCF("test", "test", field_args={"required": True})
    schema = Schema.from_dict({"test": cf.ui_field})()

    with pytest.raises(ValidationError):
        schema.load({})



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021-2024 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Transformers module."""

from abc import ABC, abstractmethod

from lxml import etree

from .errors import TransformerError
from .xml import etree_to_dict


class BaseTransformer(ABC):
    """Base transformer."""

    @abstractmethod
    def apply(self, stream_entry, *args, **kwargs):
        """Applies the transformation to the entry.

        :returns: A StreamEntry. The transformed entry.
                  Raises TransformerError in case of errors.
        """
        pass


class XMLTransformer(BaseTransformer):
    """XML transformer."""

    def __init__(self, root_element=None, *args, **kwargs):
        """Initializes the transformer."""
        self.root_element = root_element
        super().__init__(*args, **kwargs)

    @classmethod
    def _xml_to_etree(cls, xml):
        """Converts XML to a lxml etree."""
        return etree.HTML(xml)

    def apply(self, stream_entry, **kwargs):
        """Applies the transformation to the stream entry.

        Requires the root element to be named "record".
        """
        xml_tree = self._xml_to_etree(stream_entry.entry)
        xml_dict = etree_to_dict(xml_tree)["html"]["body"]

        if self.root_element:
            record = xml_dict.get(self.root_element)
            if not record:
                raise TransformerError(
                    f"Root element '{self.root_element}' not found in XML entry."
                )
        else:
            record = xml_dict

        stream_entry.entry = record
        return stream_entry



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022-2024 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""EuroSciVoc subjects datastreams, readers, transformers, and writers."""

import io
from collections import namedtuple

import requests
from rdflib import OWL, RDF, Graph, Namespace

from invenio_vocabularies.config import SUBJECTS_EUROSCIVOC_FILE_URL
from invenio_vocabularies.datastreams.readers import BaseReader
from invenio_vocabularies.datastreams.transformers import BaseTransformer


class EuroSciVocSubjectsHTTPReader(BaseReader):
    """Reader class to fetch and process EuroSciVoc RDF data."""

    def __init__(self, origin=None, mode="r", since=None, *args, **kwargs):
        """Initialize the reader with the data source.

        :param origin: The URL from which to fetch the RDF data.
        :param mode: Mode of operation (default is 'r' for reading).
        """
        self.origin = origin or SUBJECTS_EUROSCIVOC_FILE_URL
        super().__init__(origin=origin, mode=mode, *args, **kwargs)

    def _iter(self, rdf_graph):
        """Iterate over the RDF graph, yielding one subject at a time.

        :param rdf_graph: The RDF graph to process.
        :yield: Subject and graph to be transformed.
        """
        SKOS_CORE = Namespace("http://www.w3.org/2004/02/skos/core#")

        for subject, _, _ in rdf_graph.triples((None, RDF.type, SKOS_CORE.Concept)):
            yield {"subject": subject, "rdf_graph": rdf_graph}

    def read(self, item=None, *args, **kwargs):
        """Fetch and process the EuroSciVoc RDF data, yielding it one subject at a time.

        :param item: The RDF data provided as bytes (optional).
        :yield: Processed EuroSciVoc subject data.
        """
        if item:
            raise NotImplementedError(
                "EuroSciVocSubjectsHTTPReader does not support being chained after another reader"
            )
        # Fetch the RDF data from the specified origin URL
        response = requests.get(self.origin)
        response.raise_for_status()

        # Treat the response content as a file-like object
        rdf_data = io.BytesIO(response.content)

        # Parse the RDF data into a graph
        rdf_graph = Graph()
        rdf_graph.parse(rdf_data, format="xml")

        # Yield each processed subject from the RDF graph
        yield from self._iter(rdf_graph)


class EuroSciVocSubjectsTransformer(BaseTransformer):
    """Transformer class to convert EuroSciVoc RDF data to a dictionary format."""

    SKOS_CORE = Namespace("http://www.w3.org/2004/02/skos/core#")
    SPLITCHAR = ","

    def _get_notation(self, subject, rdf_graph):
        """Extract the numeric notation for a subject."""
        for _, _, notation in rdf_graph.triples(
            (subject, self.SKOS_CORE.notation, None)
        ):
            if str(notation).isdigit():
                return str(notation)
        return None

    def _get_labels(self, subject, rdf_graph):
        """Extract prefLabel and altLabel languages for a subject."""
        labels = {
            label.language: label.value.capitalize()
            for _, _, label in rdf_graph.triples(
                (subject, self.SKOS_CORE.prefLabel, None)
            )
        }
        if "en" not in labels:
            for _, _, label in rdf_graph.triples(
                (subject, self.SKOS_CORE.altLabel, None)
            ):
                labels.setdefault(label.language, label.value.capitalize())
        return labels

    def _find_parents(self, subject, rdf_graph):
        """Find parent notations."""
        parents = []

        # Traverse the broader hierarchy
        for broader in rdf_graph.transitive_objects(subject, self.SKOS_CORE.broader):
            if broader != subject:  # Ensure we don't include the current subject
                parent_notation = self._get_notation(broader, rdf_graph)
                if parent_notation:
                    parents.append(parent_notation)

        return parents

    def _transform_entry(self, subject, rdf_graph):
        """Transform an entry to the required dictionary format."""
        # Get subject notation with euroscivoc prefix
        notation = self._get_notation(subject, rdf_graph)
        id = f"euroscivoc:{notation}" if notation else None
        # Get labels for the current subject
        labels = self._get_labels(subject, rdf_graph)
        # Join parent notations with SPLITCHAR separator and add euroscivoc prefix
        parents = self.SPLITCHAR.join(
            f"euroscivoc:{n}" for n in reversed(self._find_parents(subject, rdf_graph))
        )
        # Create identifiers list
        identifiers = [{"scheme": "url", "identifier": str(subject)}]

        return {
            "id": id,
            "scheme": "EuroSciVoc",
            "subject": labels.get("en", "").capitalize(),
            "title": labels,
            "props": {"parents": parents} if parents else {},
            "identifiers": identifiers,
        }

    def apply(self, stream_entry, *args, **kwargs):
        """Transform a stream entry to the required dictionary format.

        :param stream_entry: The entry to be transformed, which includes the subject and the RDF graph.
        :return: The transformed stream entry.
        """
        # Apply transformations
        entry_data = self._transform_entry(
            stream_entry.entry["subject"], stream_entry.entry["rdf_graph"]
        )
        stream_entry.entry = entry_data
        return stream_entry


# Configuration for datastream readers, transformers, and writers
VOCABULARIES_DATASTREAM_READERS = {"euroscivoc-reader": EuroSciVocSubjectsHTTPReader}

VOCABULARIES_DATASTREAM_WRITERS = {}

VOCABULARIES_DATASTREAM_TRANSFORMERS = {
    "euroscivoc-transformer": EuroSciVocSubjectsTransformer
}

DATASTREAM_CONFIG = {
    "readers": [
        {
            "type": "euroscivoc-reader",
        }
    ],
    "transformers": [{"type": "euroscivoc-transformer"}],
    "writers": [
        {
            "type": "subjects-service",
        }
    ],
}



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2021-2024 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Test the affiliation vocabulary resource."""

import json
from copy import deepcopy

import pytest

from invenio_vocabularies.contrib.affiliations.api import Affiliation


@pytest.fixture(scope="module")
def prefix():
    """API prefix."""
    return "affiliations"


@pytest.fixture()
def example_affiliation(
    app, db, search_clear, identity, service, affiliation_full_data
):
    """Example affiliation."""
    aff = service.create(identity, affiliation_full_data)
    Affiliation.index.refresh()  # Refresh the index

    return aff


def test_affiliations_invalid(client, h, prefix):
    """Test invalid type."""
    # invalid type
    res = client.get(f"{prefix}/invalid", headers=h)
    assert res.status_code == 404


def test_affiliations_forbidden(
    client, h, prefix, example_affiliation, affiliation_full_data
):
    """Test invalid type."""
    # invalid type
    affiliation_full_data_too = deepcopy(affiliation_full_data)
    affiliation_full_data_too["id"] = "other"
    res = client.post(
        f"{prefix}", headers=h, data=json.dumps(affiliation_full_data_too)
    )
    assert res.status_code == 403

    res = client.put(
        f"{prefix}/01ggx4157", headers=h, data=json.dumps(affiliation_full_data)
    )
    assert res.status_code == 403

    res = client.delete(f"{prefix}/01ggx4157")
    assert res.status_code == 403


def test_affiliations_get(client, example_affiliation, h, prefix):
    """Test the endpoint to retrieve a single item."""
    id_ = example_affiliation.id

    res = client.get(f"{prefix}/{id_}", headers=h)
    assert res.status_code == 200
    assert res.json["id"] == id_
    # Test links
    assert res.json["links"] == {
        "self": "https://127.0.0.1:5000/api/affiliations/01ggx4157"
    }


def test_affiliations_search(client, example_affiliation, h, prefix):
    """Test a successful search."""
    res = client.get(prefix, headers=h)

    assert res.status_code == 200
    assert res.json["hits"]["total"] == 1
    assert res.json["sortBy"] == "name"


def _create_affiliations(service, identity):
    """Create dummy affiliations with similar names/acronyms/titles."""
    affiliations = [
        {
            "acronym": "CERN",
            "id": "cern",
            "name": "CERN",
            "title": {
                "en": "European Organization for Nuclear Research",
                "fr": "Conseil Européen pour la Recherche Nucléaire",
            },
        },
        {"acronym": "OTHER", "id": "other", "name": "OTHER", "title": {"en": "CERN"}},
        {
            "acronym": "CERT",
            "id": "cert",
            "name": "CERT",
            "title": {
                "en": "Computer Emergency Response Team",
                "fr": "Équipe d'Intervention d'Urgence Informatique",
            },
        },
        {
            "acronym": "NU",
            "id": "nu",
            "name": "Northwestern University",
            "title": {
                "en": "Northwestern University",
            },
        },
    ]
    for aff in affiliations:
        service.create(identity, aff)

    Affiliation.index.refresh()  # Refresh the index


def test_affiliations_prefix_search(
    app, db, search_clear, identity, service, client, h, prefix
):
    """Test a successful search."""
    _create_affiliations(service, identity)

    # Should show 1 result
    res = client.get(f"{prefix}?q=uni", headers=h)
    assert res.status_code == 200
    assert res.json["hits"]["total"] == 1


def test_affiliations_suggest_sort(
    app, db, search_clear, identity, service, client, h, prefix
):
    """Test a successful search."""
    _create_affiliations(service, identity)

    # Should show 3 results, but id=cern as first due to acronym/name, CERT due to fuzziness
    res = client.get(f"{prefix}?suggest=CERN", headers=h)
    assert res.status_code == 200
    assert res.json["hits"]["total"] == 3
    assert res.json["hits"]["hits"][0]["id"] == "cern"
    assert res.json["hits"]["hits"][1]["id"] == "cert"
    assert res.json["hits"]["hits"][2]["id"] == "other"

    # Should show 1 result, accent
    res = client.get(f"{prefix}?suggest=europeen", headers=h)
    assert res.status_code == 200
    assert res.json["hits"]["total"] == 1
    assert res.json["hits"]["hits"][0]["id"] == "cern"

    # Should show 2 results, but id=nu as first due to acronym
    res = client.get(f"{prefix}?suggest=NU", headers=h)
    assert res.status_code == 200
    assert res.json["hits"]["total"] == 2
    assert res.json["hits"]["hits"][0]["id"] == "nu"
    assert res.json["hits"]["hits"][1]["id"] == "cern"  # due to nucleaire



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022-2024 CERN.
#
# Invenio-RDM-Records is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Custom Fields for InvenioRDM."""

from .subject import SUBJECT_FIELDS, SUBJECT_FIELDS_UI
from .vocabulary import VocabularyCF

__all__ = [
    "VocabularyCF",
    "SUBJECT_FIELDS_UI" "SUBJECT_FIELDS",
]



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Funders vocabulary."""

from .resources import FundersResource, FundersResourceConfig
from .services import FundersService, FundersServiceConfig

__all__ = [
    "FundersResource",
    "FundersResourceConfig",
    "FundersService",
    "FundersServiceConfig",
]



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2024 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Invenio module for managing vocabularies."""

from .ext import InvenioVocabularies

__version__ = "6.2.0"

__all__ = ("__version__", "InvenioVocabularies")



# -*- coding: utf-8 -*-
#
# Copyright (C) 2024 University of Münster.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Names datastreams, transformers, writers and readers."""

from invenio_access.permissions import system_identity
from invenio_i18n import lazy_gettext as _

from ...datastreams.writers import ServiceWriter
from .euroscivoc import datastreams as euroscivoc_datastreams
from .mesh import datastreams as mesh_datastreams


class SubjectsServiceWriter(ServiceWriter):
    """Subjects Service Writer."""

    def __init__(self, *args, **kwargs):
        """Constructor."""
        service_or_name = kwargs.pop("service_or_name", "subjects")
        super().__init__(service_or_name=service_or_name, *args, **kwargs)

    def _entry_id(self, entry):
        """Get the id from an entry."""
        return entry["id"]


VOCABULARIES_DATASTREAM_READERS = {
    **mesh_datastreams.VOCABULARIES_DATASTREAM_READERS,
    **euroscivoc_datastreams.VOCABULARIES_DATASTREAM_READERS,
}
"""Subjects Data Streams readers."""

VOCABULARIES_DATASTREAM_TRANSFORMERS = {
    **mesh_datastreams.VOCABULARIES_DATASTREAM_TRANSFORMERS,
    **euroscivoc_datastreams.VOCABULARIES_DATASTREAM_TRANSFORMERS,
}
"""Subjects Data Streams transformers."""

VOCABULARIES_DATASTREAM_WRITERS = {
    "subjects-service": SubjectsServiceWriter,
    **mesh_datastreams.VOCABULARIES_DATASTREAM_WRITERS,
    **euroscivoc_datastreams.VOCABULARIES_DATASTREAM_WRITERS,
}
"""Subjects Data Streams writers."""

DATASTREAM_CONFIG = {
    "readers": [
        {"type": "yaml"},
    ],
    "writers": [
        {
            "type": "subjects-service",
        }
    ],
}
"""Data Stream configuration."""



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021-2024 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Names data streams tests."""

from copy import deepcopy
from unittest.mock import patch

import pytest
from invenio_access.permissions import system_identity
from invenio_records_resources.proxies import current_service_registry

from invenio_vocabularies.contrib.names.api import Name
from invenio_vocabularies.contrib.names.datastreams import (
    NamesServiceWriter,
    OrcidHTTPReader,
    OrcidTransformer,
)
from invenio_vocabularies.datastreams import StreamEntry
from invenio_vocabularies.datastreams.errors import TransformerError, WriterError


@pytest.fixture(scope="function")
def name_full_data():
    """Full name data."""
    return {
        "id": "0000-0001-8135-3489",
        "given_name": "Lars Holm",
        "family_name": "Nielsen",
        "name": "Nielsen, Lars Holm",
        "identifiers": [
            {
                "scheme": "orcid",
                "identifier": "0000-0001-8135-3489",  # normalized after create
            }
        ],
        "affiliations": [{"name": "CERN"}],
    }


@pytest.fixture(scope="module")
def expected_from_xml():
    return {
        "id": "0000-0001-8135-3489",
        "given_name": "Lars Holm",
        "family_name": "Nielsen",
        "identifiers": [{"scheme": "orcid", "identifier": "0000-0001-8135-3489"}],
        "affiliations": [{"name": "CERN"}],
    }


XML_ENTRY_DATA = bytes(
    '<?xml version="1.0" encoding="UTF-8" standalone="yes"?>\n'
    '<record:record path="/0000-0001-8135-3489">\n'
    "    <common:orcid-identifier>\n"
    "        <common:uri>https://orcid.org/0000-0001-8135-3489</common:uri>\n"  # noqa
    "        <common:path>0000-0001-8135-3489</common:path>\n"
    "        <common:host>orcid.org</common:host>\n"
    "    </common:orcid-identifier>\n"
    '    <person:person path="/0000-0001-8135-3489/person">\n'
    '        <person:name visibility="public" path="0000-0001-8135-3489">\n'  # noqa
    "            <personal-details:given-names>Lars Holm</personal-details:given-names>"  # noqa
    "            <personal-details:family-name>Nielsen</personal-details:family-name>\n"  # noqa
    "        </person:name>\n"
    '        <external-identifier:external-identifiers path="/0000-0001-8135-3489/external-identifiers"/>\n'  # noqa
    "    </person:person>\n"
    '    <activities:activities-summary path="/0000-0001-8135-3489/activities">\n'  # noqa
    '       <activities:employments path="/0000-0001-8135-3489/employments">\n'  # noqa
    "           <employments:affiliation-group>\n"
    "               <employments:employment-summary>\n"
    "                   <employment-summary:organization>\n"
    "                       <organization:name>CERN</organization:name>\n"
    "                   </employment-summary:organization>\n"
    "               </employments:employment-summary>\n"
    "           </employments:affiliation-group>\n"
    "       </activities:employments>\n"
    "    </activities:activities-summary>\n"
    "</record:record>\n",
    encoding="raw_unicode_escape",
)


@pytest.fixture(scope="function")
def bytes_xml_data():
    # simplified version of an XML file of the ORCiD dump
    return XML_ENTRY_DATA


NAMES_TEST = {
    "Lars Holm, Nielsen": True,
    "Lars Nielsen:)": False,
    "Lars Nielsen-H": True,
    "Lars Nielsen 👍": False,
    "Lars Nielsen (CERN)": True,
}


@pytest.fixture(scope="function")
def dict_xml_entry():
    return StreamEntry(
        {
            "orcid-identifier": {
                "uri": "https://orcid.org/0000-0001-8135-3489",
                "path": "0000-0001-8135-3489",
                "host": "orcid.org",
            },
            "person": {
                "name": {
                    "given-names": "Lars Holm",
                    "family-name": "Nielsen",
                    "@visibility": "public",
                    "@path": "0000-0001-8135-3489",
                },
                "external-identifiers": {
                    "@path": "/0000-0001-8135-3489/external-identifiers"
                },
                "@path": "/0000-0001-8135-3489/person",
            },
            "activities-summary": {
                "employments": {
                    "affiliation-group": {
                        "employment-summary": {"organization": {"name": "CERN"}}
                    },
                    "@path": "/0000-0001-8135-3489/employments",
                },
                "@path": "/0000-0001-8135-3489/activities",
            },
            "@path": "/0000-0001-8135-3489",
        }
    )


def test_orcid_transformer(dict_xml_entry, expected_from_xml):
    transformer = OrcidTransformer()
    assert expected_from_xml == transformer.apply(dict_xml_entry).entry


@pytest.mark.parametrize("name,is_valid_name", NAMES_TEST.items())
def test_orcid_transformer_different_names(dict_xml_entry, name, is_valid_name):
    transformer = OrcidTransformer()
    val = deepcopy(dict_xml_entry)
    if is_valid_name:
        assert transformer.apply(val).entry
    else:
        with pytest.raises(TransformerError):
            val.entry["person"]["name"]["given-names"] = name
            val.entry["person"]["name"]["family-name"] = ""
            transformer.apply(val)
        with pytest.raises(TransformerError):
            val.entry["person"]["name"]["given-names"] = ""
            val.entry["person"]["name"]["family-name"] = name
            transformer.apply(val)


@pytest.mark.parametrize("name", NAMES_TEST.keys())
def test_orcid_transformer_different_names_no_regex(dict_xml_entry, name):
    transformer = OrcidTransformer(names_exclude_regex=None)
    val = deepcopy(dict_xml_entry)
    val.entry["person"]["name"]["given-names"] = name
    val.entry["person"]["name"]["family-name"] = ""
    assert transformer.apply(val).entry


class MockResponse:
    content = XML_ENTRY_DATA
    status_code = 200


@patch("requests.get", side_effect=lambda url, headers: MockResponse())
def test_orcid_http_reader(_, bytes_xml_data):
    reader = OrcidHTTPReader(id="0000-0001-8135-3489")
    results = []
    for entry in reader.read():
        results.append(entry)

    assert len(results) == 1
    assert bytes_xml_data == results[0]


def test_names_service_writer_create(app, db, search_clear, name_full_data):
    writer = NamesServiceWriter()
    record = writer.write(StreamEntry(name_full_data))
    record = record.entry.to_dict()

    assert dict(record, **name_full_data) == record


def test_names_service_writer_duplicate(app, db, search_clear, name_full_data):
    writer = NamesServiceWriter()
    _ = writer.write(stream_entry=StreamEntry(name_full_data))
    Name.index.refresh()  # refresh index to make changes live
    with pytest.raises(WriterError) as err:
        writer.write(stream_entry=StreamEntry(name_full_data))

    expected_error = [f"Vocabulary entry already exists: {name_full_data}"]
    assert expected_error in err.value.args


def test_names_service_writer_update_existing(app, db, search_clear, name_full_data):
    # create vocabulary
    writer = NamesServiceWriter(update=True)
    name = writer.write(stream_entry=StreamEntry(name_full_data))
    Name.index.refresh()  # refresh index to make changes live
    # update vocabulary
    updated_name = deepcopy(name_full_data)
    updated_name["given_name"] = "Pablo"
    updated_name["family_name"] = "Panero"
    updated_name["name"] = "Panero, Pablo"
    # check changes vocabulary
    _ = writer.write(stream_entry=StreamEntry(updated_name))
    service = current_service_registry.get("names")
    record = service.read(system_identity, name.entry.id)
    record = record.to_dict()

    # needed while the writer resolves from ES
    assert _.entry.id == name.entry.id
    assert dict(record, **updated_name) == record


def test_names_service_writer_update_non_existing(
    app, db, search_clear, name_full_data
):
    # vocabulary item not created, call update directly
    updated_name = deepcopy(name_full_data)
    updated_name
    updated_name["given_name"] = "Pablo"
    updated_name["family_name"] = "Panero"
    updated_name["name"] = "Panero, Pablo"
    # check changes vocabulary
    writer = NamesServiceWriter(update=True)
    name = writer.write(stream_entry=StreamEntry(updated_name))
    service = current_service_registry.get("names")
    record = service.read(system_identity, name.entry.id)
    record = record.to_dict()

    assert dict(record, **updated_name) == record



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022-2024 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Awards datastreams tests."""

from copy import deepcopy

import pytest
from invenio_access.permissions import system_identity

from invenio_vocabularies.contrib.awards.api import Award
from invenio_vocabularies.contrib.awards.datastreams import (
    AwardsServiceWriter,
    CORDISProjectTransformer,
    OpenAIREProjectTransformer,
)
from invenio_vocabularies.datastreams import StreamEntry
from invenio_vocabularies.datastreams.errors import WriterError
from invenio_vocabularies.datastreams.readers import XMLReader


@pytest.fixture(scope="function")
def dict_award_entry():
    return StreamEntry(
        {
            "acronym": "TA",
            "code": "0751743",
            "enddate": "2010-09-30",
            "funding": [
                {
                    "fundingStream": {
                        "description": "Directorate for Geosciences - Division of "
                        "Ocean Sciences",
                        "id": "NSF::GEO/OAD::GEO/OCE",
                    },
                    "jurisdiction": "US",
                    "name": "National Science Foundation",
                    "shortName": "NSF",
                }
            ],
            "h2020programme": [],
            "id": "nsf_________::3eb1b4f6d6e251a19f9fdeed2aab88d8",
            "openaccessmandatefordataset": False,
            "openaccessmandateforpublications": False,
            "startdate": "2008-04-01",
            "subject": ["Oceanography"],
            "title": "Test title",
            "websiteurl": "https://test.com",
        }
    )


@pytest.fixture(scope="function")
def dict_award_entry_ec():
    """Full award data."""
    return StreamEntry(
        {
            "acronym": "TS",
            "code": "129123",
            "enddate": "2025-12-31",
            "funding": [
                {
                    "fundingStream": {
                        "description": "Test stream",
                        "id": "TST::test::test",
                    },
                    "jurisdiction": "GR",
                    "name": "Test Name",
                    "shortName": "TST",
                }
            ],
            "h2020programme": [],
            "id": "40|corda__h2020::000000000000000000",
            "openaccessmandatefordataset": False,
            "openaccessmandateforpublications": False,
            "startdate": "2008-04-01",
            "subject": ["Oceanography"],
            "title": "Test title",
            "websiteurl": "https://test.com",
        }
    )


@pytest.fixture(scope="function")
def expected_from_award_json():
    return {
        "id": "021nxhr62::0751743",
        "identifiers": [{"identifier": "https://test.com", "scheme": "url"}],
        "number": "0751743",
        "program": "NSF::GEO/OAD::GEO/OCE",
        "title": {"en": "Test title"},
        "funder": {"id": "021nxhr62"},
        "acronym": "TA",
    }


@pytest.fixture(scope="function")
def expected_from_award_json_ec():
    return {
        "id": "00k4n6c32::129123",
        "identifiers": [
            {"identifier": "https://cordis.europa.eu/projects/129123", "scheme": "url"}
        ],
        "number": "129123",
        "title": {"en": "Test title"},
        "funder": {"id": "00k4n6c32"},
        "acronym": "TS",
        "program": "TST::test::test",
    }


CORDIS_PROJECT_XML = bytes(
    """      
<project
	xmlns="http://cordis.europa.eu">
	<language>en</language>
	<availableLanguages readOnly="true">en</availableLanguages>
	<rcn>264556</rcn>
	<id>101117736</id>
	<acronym>Time2SWITCH</acronym>
	<identifiers>
		<grantDoi>10.3030/101117736</grantDoi>
	</identifiers>
	<relations>
		<associations>
			<organization type="coordinator" source="corda" order="1" ecContribution="1496795" terminated="false" sme="false" netEcContribution="1496795" totalCost="1496795">
				<availableLanguages readOnly="true">en</availableLanguages>
				<rcn>1908489</rcn>
				<id>999979888</id>
				<vatNumber>ATU37675002</vatNumber>
				<legalName>TECHNISCHE UNIVERSITAET WIEN</legalName>
				<shortName>TU WIEN</shortName>
				<departmentName>Institute of Electrodynamics, Microwave and Circuit Engineer</departmentName>
				<relations>
					<categories>
						<category classification="organizationActivityType" type="relatedOrganizationActivityType">
							<language>en</language>
							<availableLanguages readOnly="true">de,en,es,fr,it,pl</availableLanguages>
							<code>HES</code>
							<title>Higher or Secondary Education Establishments</title>
							<displayCode readOnly="true">/Higher or Secondary Education Establishments</displayCode>
						</category>
					</categories>
				</relations>
			</organization>
		</associations>
		<categories>
			<category classification="euroSciVoc" type="isInFieldOfScience">
				<language>en</language>
				<availableLanguages readOnly="true">de,en,es,fr,it,pl</availableLanguages>
				<code>/21/39/225</code>
				<title>oncology</title>
				<displayCode readOnly="true">/medical and health sciences/clinical medicine/oncology</displayCode>
			</category>
		</categories>
	</relations>
</project>
    """,
    encoding="utf-8",
)


@pytest.fixture(scope="module")
def expected_from_cordis_project_xml():
    return {
        "id": "00k4n6c32::101117736",
        "subjects": [{"id": "euroscivoc:225"}],
        "organizations": [
            {
                "id": "999979888",
                "scheme": "pic",
                "organization": "TECHNISCHE UNIVERSITAET WIEN",
            }
        ],
    }


def test_awards_transformer(app, dict_award_entry, expected_from_award_json):
    """Test the OpenAIREProjectTransformer's output against expected award data."""
    transformer = OpenAIREProjectTransformer()
    assert expected_from_award_json == transformer.apply(dict_award_entry).entry


def test_awards_service_writer_create(
    app, search_clear, example_funder_ec, award_full_data
):
    """Verify creation of an award record and match it with expected data."""
    awards_writer = AwardsServiceWriter()
    award_rec = awards_writer.write(StreamEntry(award_full_data))
    award_dict = award_rec.entry.to_dict()

    award_full_data["funder"]["name"] = example_funder_ec["name"]
    assert dict(award_dict, **award_full_data) == award_dict

    # not-ideal cleanup
    award_rec.entry._record.delete(force=True)


def test_awards_funder_id_not_exist(
    app, search_clear, example_funder, example_funder_ec, award_full_data_invalid_id
):
    """Ensure writing an award with an invalid funder ID raises an error."""
    awards_writer = AwardsServiceWriter()
    with pytest.raises(WriterError) as err:
        awards_writer.write(StreamEntry(award_full_data_invalid_id))
    expected_error = [
        {
            "InvalidRelationValue": "Invalid value {funder_id}.".format(
                funder_id=award_full_data_invalid_id.get("funder").get("id")
            )
        }
    ]

    assert expected_error in err.value.args


def test_awards_funder_id_not_exist_no_funders(
    app, search_clear, award_full_data_invalid_id
):
    """Check error on writing an award with no valid funders."""
    awards_writer = AwardsServiceWriter()
    with pytest.raises(WriterError) as err:
        awards_writer.write(StreamEntry(award_full_data_invalid_id))
    expected_error = [
        {
            "InvalidRelationValue": "Invalid value {funder_id}.".format(
                funder_id=award_full_data_invalid_id.get("funder").get("id")
            )
        }
    ]

    assert expected_error in err.value.args


def test_awards_transformer_ec_functionality(
    app,
    search_clear,
    dict_award_entry,
    dict_award_entry_ec,
    expected_from_award_json,
    expected_from_award_json_ec,
):
    """Test transformer output for standard and EC-specific awards."""
    transformer = OpenAIREProjectTransformer()

    assert expected_from_award_json == transformer.apply(dict_award_entry).entry
    assert expected_from_award_json_ec == transformer.apply(dict_award_entry_ec).entry


def test_awards_service_writer_duplicate(
    app, search_clear, example_funder_ec, award_full_data
):
    """Verify error on attempting to create a duplicate award."""
    writer = AwardsServiceWriter()
    award_rec = writer.write(stream_entry=StreamEntry(award_full_data))
    Award.index.refresh()  # refresh index to make changes live
    with pytest.raises(WriterError) as err:
        writer.write(stream_entry=StreamEntry(award_full_data))

    expected_error = [f"Vocabulary entry already exists: {award_full_data}"]
    assert expected_error in err.value.args

    # not-ideal cleanup
    award_rec.entry._record.delete(force=True)


def test_awards_service_writer_update_existing(
    app, search_clear, example_funder_ec, award_full_data, service
):
    """Check updating an existing award record with new data."""
    writer = AwardsServiceWriter(update=True)
    orig_award_rec = writer.write(stream_entry=StreamEntry(award_full_data))
    Award.index.refresh()  # refresh index to make changes live
    updated_award = deepcopy(award_full_data)
    updated_award["title"] = {"en": "New Test title"}
    _ = writer.write(stream_entry=StreamEntry(updated_award))
    award_rec = service.read(system_identity, orig_award_rec.entry.id)
    award_dict = award_rec.to_dict()

    updated_award["funder"]["name"] = example_funder_ec["name"]
    assert _.entry.id == orig_award_rec.entry.id
    assert dict(award_dict, **updated_award) == award_dict

    # not-ideal cleanup
    award_rec._record.delete(force=True)


def test_awards_service_writer_update_non_existing(
    app, search_clear, example_funder_ec, award_full_data, service
):
    """Test updating a non-existing award, creating a new record."""
    updated_award = deepcopy(award_full_data)
    updated_award["title"] = {"en": "New Test title"}
    writer = AwardsServiceWriter(update=True)
    award_rec = writer.write(stream_entry=StreamEntry(updated_award))
    award_rec = service.read(system_identity, award_rec.entry.id)
    award_dict = award_rec.to_dict()

    updated_award["funder"]["name"] = example_funder_ec["name"]
    assert dict(award_dict, **updated_award) == award_dict

    # not-ideal cleanup
    award_rec._record.delete(force=True)


def test_awards_cordis_transformer(expected_from_cordis_project_xml):
    """Validate transformation of CORDIS project XML to expected format."""
    reader = XMLReader()
    award = next(reader.read(CORDIS_PROJECT_XML))

    cordis_transformer = CORDISProjectTransformer()
    transformed_award_data = cordis_transformer.apply(StreamEntry(award["project"]))

    assert transformed_award_data.entry == expected_from_cordis_project_xml



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Awards API tests."""

from copy import deepcopy
from functools import partial

import pytest
from invenio_indexer.api import RecordIndexer
from invenio_search import current_search_client
from jsonschema import ValidationError as SchemaValidationError

from invenio_vocabularies.contrib.awards.api import Award


@pytest.fixture()
def search_get():
    """Get a document from an index."""
    return partial(current_search_client.get, Award.index._name)


@pytest.fixture()
def example_award(db, award_full_data, example_funder_ec):
    """Example award."""
    api_award = deepcopy(award_full_data)
    pid = api_award.pop("id")  # at API level it's passed as an arg
    awa = Award.create(api_award, pid=pid)
    awa.commit()
    db.session.commit()
    return awa


def test_award_schema_validation(app, example_award):
    """Award schema validation."""
    # valid data
    awa = example_award

    assert awa.schema == "local://awards/award-v1.0.0.json"
    assert awa.pid
    assert awa.id

    # invalid data
    examples = [
        # title are objects of key/string.
        {"title": "not a dict"},
        {"title": {"en": 123}},
        # identifiers are objects of key/string.
        {"identifiers": "03yrm5c26"},
        {"identifiers": ["03yrm5c26"]},
        {"identifiers": {"03yrm5c26"}},
        # number must be a string
        {"number": 123},
        # funder must be an object
        {"funder": 123},
    ]

    for ex in examples:
        pytest.raises(SchemaValidationError, Award.create, ex)


def test_award_indexing(app, example_award, indexer, search_get):
    """Test indexing of an award."""
    # Index document in ES
    assert indexer.index(example_award)["result"] == "created"

    # Retrieve document from ES
    data = search_get(id=example_award.id)

    # Loads the ES data and compare
    awa = Award.loads(data["_source"])
    assert awa == example_award
    assert awa.id == example_award.id
    assert awa.pid.pid_value == example_award.pid.pid_value
    assert awa.revision_id == example_award.revision_id
    assert awa.created == example_award.created
    assert awa.updated == example_award.updated


def test_award_pid(app, example_award):
    """Test award pid creation."""
    awa = example_award

    assert awa.pid.pid_value == "755021"
    assert Award.pid.resolve("755021") == example_award



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 Northwestern University.
# Copyright (C) 2021-2024 CERN.
# Copyright (C) 2024 University of Münster.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Subjects schema."""

from functools import partial

from invenio_i18n import get_locale
from marshmallow import Schema, fields, pre_load
from marshmallow_utils.fields import IdentifierSet, SanitizedUnicode
from marshmallow_utils.schemas import IdentifierSchema

from ...services.schema import (
    BaseVocabularySchema,
    ContribVocabularyRelationSchema,
    i18n_strings,
)
from .config import subject_schemes


class SubjectSchema(BaseVocabularySchema):
    """Service schema for subjects."""

    # id in BaseRecordSchema is not required, but I don't see why it shouldn't
    # be, while scheme and subject are required. So I am making it required
    # here.
    id = SanitizedUnicode(required=True)
    scheme = SanitizedUnicode(required=True)
    subject = SanitizedUnicode(required=True)
    title = i18n_strings
    props = fields.Dict(keys=SanitizedUnicode(), values=SanitizedUnicode())
    identifiers = IdentifierSet(
        fields.Nested(
            partial(
                IdentifierSchema,
                allowed_schemes=subject_schemes,
                identifier_required=False,
            )
        )
    )
    synonyms = fields.List(SanitizedUnicode())

    @pre_load
    def add_subject_from_title(self, data, **kwargs):
        """Add subject from title if not present."""
        locale = get_locale().language
        if "subject" not in data:
            data["subject"] = data["title"].get(locale) or data["title"].values()[0]
        return data


class SubjectRelationSchema(ContribVocabularyRelationSchema):
    """Schema to define an optional subject relation in another schema."""

    ftf_name = "subject"
    parent_field_name = "subjects"
    subject = SanitizedUnicode()
    scheme = SanitizedUnicode(dump_only=True)
    title = fields.Dict(dump_only=True)
    props = fields.Dict(dump_only=True)
    identifiers = IdentifierSet(
        fields.Nested(
            partial(
                IdentifierSchema,
                allowed_schemes=subject_schemes,
                identifier_required=False,
            )
        )
    )
    synonyms = fields.List(SanitizedUnicode(), dump_only=True)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2021-2024 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Test the name vocabulary resource."""

import json
from copy import deepcopy

import pytest

from invenio_vocabularies.contrib.names.api import Name


@pytest.fixture(scope="module")
def prefix():
    """API prefix."""
    return "names"


@pytest.fixture()
def names_data():
    """Full name data."""
    return [
        {
            "id": "0000-0001-8135-3489",
            "given_name": "Niels John",
            "family_name": "Davidson",
            "name": "Davidson, Niels John",
            "identifiers": [
                {
                    "scheme": "orcid",
                    "identifier": "0000-0001-8135-3489",
                }
            ],
            "affiliations": [{"name": "CERN"}],
        },
        {
            "id": "0000-0002-8438-3752",
            "given_name": "Dwäyne",
            "family_name": "Johnsœn",
            "name": "Johnsœn, Dwäyne",
            "identifiers": [
                {
                    "scheme": "orcid",
                    "identifier": "0000-0002-8438-3752",
                }
            ],
            "affiliations": [{"name": "CERN"}],
        },
        {
            "id": "0000-0002-0816-7126",
            "given_name": "John",
            "family_name": "Cena",
            "name": "Cena, John",
            "identifiers": [
                {
                    "scheme": "orcid",
                    "identifier": "0000-0002-0816-7126",
                }
            ],
            "affiliations": [{"name": "WWE"}],
        },
    ]


@pytest.fixture()
def example_multiple_names(
    app, db, search_clear, identity, service, names_data, example_affiliation
):
    """Example multiple names."""
    names = []
    for i in range(len(names_data)):
        names.append(service.create(identity, names_data[i]))
        Name.index.refresh()  # Refresh the index

    return names


@pytest.fixture()
def example_name(
    app, db, search_clear, identity, service, name_full_data, example_affiliation
):
    """Example name."""
    name = service.create(identity, name_full_data)
    Name.index.refresh()  # Refresh the index

    return name


def test_non_authenticated(client, prefix, example_name, h):
    """Test non-authenticated access."""
    # Search
    res = client.get(prefix)
    assert res.status_code == 403

    id_ = example_name.id

    # Read
    res = client.get(f"{prefix}/{id_}", headers=h)
    assert res.status_code == 403

    # Resolve
    res = client.get(f"{prefix}/orcid/0000-0001-8135-3489", headers=h)
    assert res.status_code == 403


def test_names_invalid(client_with_credentials, h, prefix):
    """Test invalid type."""
    # invalid type
    res = client_with_credentials.get(f"{prefix}/invalid", headers=h)
    assert res.status_code == 404


def test_names_forbidden(client, h, prefix, example_name, name_full_data):
    """Test invalid type."""
    # invalid type
    res = client.post(f"{prefix}", headers=h, data=json.dumps(name_full_data))
    assert res.status_code == 403

    res = client.put(
        f"{prefix}/{example_name.id}", headers=h, data=json.dumps(name_full_data)
    )
    assert res.status_code == 403

    res = client.delete(f"{prefix}/{example_name.id}")
    assert res.status_code == 403


def test_names_get(client_with_credentials, example_name, h, prefix):
    """Test the endpoint to retrieve a single item."""
    id_ = example_name.id

    res = client_with_credentials.get(f"{prefix}/{id_}", headers=h)
    assert res.status_code == 200
    assert res.json["id"] == id_
    # Test links
    assert res.json["links"] == {
        "self": f"https://127.0.0.1:5000/api/names/{example_name.id}"
    }


def test_names_search(client_with_credentials, example_name, h, prefix):
    """Test a successful search."""
    res = client_with_credentials.get(prefix, headers=h)

    assert res.status_code == 200
    assert res.json["hits"]["total"] == 1
    assert res.json["sortBy"] == "name"


def test_names_resolve(client_with_credentials, example_name, h, prefix):
    res = client_with_credentials.get(f"{prefix}/orcid/0000-0001-8135-3489", headers=h)
    assert res.status_code == 200
    assert res.json["id"] == example_name.id

    res = client_with_credentials.get(f"{prefix}/orcid/0000-0002-5082-6404", headers=h)
    assert res.status_code == 404


def test_names_suggest_sort(client_with_credentials, example_multiple_names, h, prefix):
    """Test a successful search."""

    # With typo
    res = client_with_credentials.get(f"{prefix}?suggest=davisson", headers=h)
    assert res.status_code == 200
    assert res.json["hits"]["total"] == 1
    assert res.json["hits"]["hits"][0]["name"] == "Davidson, Niels John"

    # With accent
    res = client_with_credentials.get(f"{prefix}?suggest=dwayne", headers=h)
    assert res.status_code == 200
    assert res.json["hits"]["total"] == 1
    assert res.json["hits"]["hits"][0]["name"] == "Johnsœn, Dwäyne"

    res = client_with_credentials.get(
        f"{prefix}?suggest=Dw%C3%A4yne", headers=h
    )  # Dwäyne
    assert res.status_code == 200
    assert res.json["hits"]["total"] == 1
    assert res.json["hits"]["hits"][0]["name"] == "Johnsœn, Dwäyne"

    # With incomplete
    res = client_with_credentials.get(f"{prefix}?suggest=joh", headers=h)
    assert res.status_code == 200
    assert res.json["hits"]["total"] == 3

    # With affiliation
    res = client_with_credentials.get(f"{prefix}?suggest=john%20wwe", headers=h)
    assert res.status_code == 200
    assert (
        res.json["hits"]["total"] == 3
    )  # Will find 3 johns but WWE affiliation should be at the top
    assert res.json["hits"]["hits"][0]["name"] == "Cena, John"
    assert res.json["hits"]["hits"][0]["affiliations"][0]["name"] == "WWE"



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Names API tests."""

from copy import deepcopy
from functools import partial

import pytest
from invenio_indexer.api import RecordIndexer
from invenio_records.systemfields.relations.errors import InvalidRelationValue
from invenio_search import current_search_client
from jsonschema import ValidationError as SchemaValidationError

from invenio_vocabularies.contrib.names.api import Name


@pytest.fixture()
def search_get():
    """Get a document from an index."""
    return partial(current_search_client.get, Name.index._name)


@pytest.fixture()
def indexer():
    """Indexer instance with correct Record class."""
    return RecordIndexer(
        record_cls=Name,
        record_to_index=lambda r: (r.__class__.index._name, "_doc"),
    )


@pytest.fixture()
def example_name(db, name_full_data, example_affiliation):
    """Example name."""
    api_name = deepcopy(name_full_data)
    pid = api_name.pop("id")  # at API level it's passed as an arg
    name = Name.create(api_name, pid=pid)
    name.commit()
    db.session.commit()
    return name


def test_name_schema_validation(app, db, example_name):
    """Name schema validation."""
    # valid data
    name = example_name

    assert name.schema == "local://names/name-v1.0.0.json"
    assert name.pid
    assert name.id

    # invalid data
    examples = [
        # identifiers are objects of key/string.
        {"name": "Doe, John", "identifiers": "0000-0001-8135-3489"},
        {"name": "Doe, John", "identifiers": ["0000-0001-8135-3489"]},
        {"name": "Doe, John", "identifiers": {"0000-0001-8135-3489"}},
        # names must be a string
        {"name": 123},
        {"family_name": 123, "given_name": "John"},
        {"family_name": "Doe", "given_name": 123},
        # affiliations are objects of key/string.
        {"name": "Doe, John", "affiliations": "cern"},
        {"name": "Doe, John", "affiliations": ["cern"]},
        {"name": "Doe, John", "affiliations": {"cern"}},
    ]

    for ex in examples:
        pytest.raises(SchemaValidationError, Name.create, ex)

    # test an affiliation that does not exist in the db
    invalid_aff = {"name": "Doe, John", "affiliations": [{"id": "invalid"}]}
    invalid_name = Name.create(invalid_aff)
    pytest.raises(InvalidRelationValue, invalid_name.commit)


def test_name_indexing(app, db, search, example_name, indexer, search_get):
    """Test indexing of a name."""
    # Index document in ES
    assert indexer.index(example_name)["result"] == "created"

    # Retrieve document from ES
    data = search_get(id=example_name.id)

    # Loads the ES data and compare
    name = Name.loads(data["_source"])
    assert name == example_name
    assert name.id == example_name.id
    assert name.revision_id == example_name.revision_id
    assert name.created == example_name.created
    assert name.updated == example_name.updated


def test_name_pid(app, db, example_name):
    """Test name pid creation."""
    assert example_name.pid.pid_value == "0000-0001-8135-3489"



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021-2022 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Fixtures tests."""

from pathlib import Path

import pytest

from invenio_vocabularies.fixtures import VocabularyFixture


def test_base_fixture(app):
    """Simulates looping through two vocabularies.

    Reports a set of errors per vocabulary. In this case each vocab has one.
    The first fails on a transformer and the second fails on a writer.
    Note that the entries are not the first one, so only the ones with errors
    are returned.
    """
    filepath = Path(__file__).parent / "vocabularies.yaml"
    fixture = VocabularyFixture(filepath=filepath)
    results = fixture.load()

    invalid_tr = next(results)
    assert len(invalid_tr) == 1
    result = invalid_tr[0]
    assert result.entry == -1
    assert ["TestTransformer: Value cannot be negative"] == result.errors

    invalid_wr = next(results)
    assert len(invalid_wr) == 1
    result = invalid_wr[0]
    assert result.entry == 2
    assert ["FailingTestWriter: 2 value found."] == result.errors

    with pytest.raises(StopIteration):
        next(results)  # stops after the two configured vocabularies



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021-2022 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Jobs module."""

import datetime
from datetime import timezone

from invenio_i18n import gettext as _
from invenio_jobs.jobs import JobType
from marshmallow import Schema, fields
from marshmallow_utils.fields import TZDateTime

from invenio_vocabularies.services.tasks import process_datastream


class ArgsSchema(Schema):
    """Schema of task input arguments."""

    since = TZDateTime(
        timezone=timezone.utc,
        format="iso",
        metadata={
            "description": _(
                "YYYY-MM-DD HH:mm format. "
                "Leave field empty if it should continue since last successful run."
            )
        },
    )
    job_arg_schema = fields.String(
        metadata={"type": "hidden"},
        dump_default="ArgsSchema",
        load_default="ArgsSchema",
    )


class ProcessDataStreamJob(JobType):
    """Generic process data stream job type."""

    arguments_schema = ArgsSchema
    task = process_datastream
    id = None


class ProcessRORAffiliationsJob(ProcessDataStreamJob):
    """Process ROR affiliations datastream registered task."""

    description = "Process ROR affiliations"
    title = "Load ROR affiliations"
    id = "process_ror_affiliations"

    @classmethod
    def default_args(cls, job_obj, since=None, **kwargs):
        """Generate default job arguments here."""
        if since is None and job_obj.last_runs["success"]:
            since = job_obj.last_runs["success"].started_at
        else:
            since = datetime.datetime.now()

        return {
            "config": {
                "readers": [
                    {
                        "args": {"since": since},
                        "type": "ror-http",
                    },
                    {"args": {"regex": "_schema_v2\\.json$"}, "type": "zip"},
                    {"type": "json"},
                ],
                "writers": [
                    {
                        "args": {
                            "writer": {
                                "type": "affiliations-service",
                                "args": {"update": True},
                            }
                        },
                        "type": "async",
                    }
                ],
                "transformers": [{"type": "ror-affiliations"}],
            }
        }



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2021 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Vocabulary data API."""

from invenio_records.dumpers import SearchDumper
from invenio_records.dumpers.indexedat import IndexedAtDumperExt
from invenio_records.systemfields import ConstantField, RelatedModelField
from invenio_records_resources.records.api import Record
from invenio_records_resources.records.systemfields import IndexField, PIDField

from .models import VocabularyMetadata, VocabularyType
from .pidprovider import VocabularyIdProvider
from .systemfields import VocabularyPIDFieldContext


class Vocabulary(Record):
    """A generic vocabulary record."""

    # Configuration
    model_cls = VocabularyMetadata

    # System fields
    # TODO: Can schema name be changed (remove localhost)
    schema = ConstantField(
        "$schema",
        "local://vocabularies/vocabulary-v1.0.0.json",
    )

    index = IndexField("vocabularies-vocabulary-v1.0.0", search_alias="vocabularies")

    #: Disable the metadata system field.
    metadata = None

    type = RelatedModelField(VocabularyType, required=True)

    pid = PIDField(
        "id",
        provider=VocabularyIdProvider,
        context_cls=VocabularyPIDFieldContext,
        create=False,
    )

    dumper = SearchDumper(
        extensions=[
            IndexedAtDumperExt(),
        ]
    )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021-2024 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Writers module."""

from abc import ABC, abstractmethod
from pathlib import Path

import yaml
from invenio_access.permissions import system_identity
from invenio_pidstore.errors import PIDAlreadyExists, PIDDoesNotExistError
from invenio_records.systemfields.relations.errors import InvalidRelationValue
from invenio_records_resources.proxies import current_service_registry
from marshmallow import ValidationError
from sqlalchemy.exc import NoResultFound

from .datastreams import StreamEntry
from .errors import WriterError
from .tasks import write_entry, write_many_entry


class BaseWriter(ABC):
    """Base writer."""

    def __init__(self, *args, **kwargs):
        """Base initialization logic."""
        # Add any base initialization here if needed
        pass

    @abstractmethod
    def write(self, stream_entry, *args, **kwargs):
        """Writes the input stream entry to the target output.

        :returns: A StreamEntry. The result of writing the entry.
                  Raises WriterException in case of errors.

        """
        pass

    @abstractmethod
    def write_many(self, stream_entries, *args, **kwargs):
        """Writes the input streams entry to the target output.

        :returns: A List of StreamEntry. The result of writing the entry.
                  Raises WriterException in case of errors.

        """
        pass


class ServiceWriter(BaseWriter):
    """Writes the entries to an RDM instance using a Service object."""

    def __init__(
        self, service_or_name, *args, identity=None, insert=True, update=False, **kwargs
    ):
        """Constructor.

        :param service_or_name: a service instance or a key of the
                                service registry.
        :param identity: access identity.
        :param insert: if True it will insert records which do not exist.
        :param update: if True it will update records if they exist.
        """
        if isinstance(service_or_name, str):
            service_or_name = current_service_registry.get(service_or_name)

        self._service = service_or_name
        self._identity = identity or system_identity
        self._insert = insert
        self._update = update

        super().__init__(*args, **kwargs)

    def _entry_id(self, entry):
        """Get the id from an entry."""
        return (entry["type"], entry["id"])

    def _resolve(self, id_):
        return self._service.read(self._identity, id_)

    def _do_update(self, entry):
        vocab_id = self._entry_id(entry)
        current = self._resolve(vocab_id)
        combined_dict = current.to_dict()

        # Update fields from entry
        for key, value in entry.items():
            if key in combined_dict:
                if isinstance(combined_dict[key], list) and isinstance(value, list):
                    combined_dict[key].extend(
                        item for item in value if item not in combined_dict[key]
                    )
                else:
                    combined_dict[key] = value
            else:
                combined_dict[key] = value

        return StreamEntry(
            self._service.update(self._identity, vocab_id, combined_dict)
        )

    def write(self, stream_entry, *args, **kwargs):
        """Writes the input entry using a given service."""
        entry = stream_entry.entry

        try:
            if self._insert:
                try:
                    return StreamEntry(self._service.create(self._identity, entry))
                except PIDAlreadyExists:
                    if not self._update:
                        raise WriterError([f"Vocabulary entry already exists: {entry}"])
                    return self._do_update(entry)
            elif self._update:
                try:
                    return self._do_update(entry)
                except (NoResultFound, PIDDoesNotExistError):
                    raise WriterError([f"Vocabulary entry does not exist: {entry}"])
            else:
                raise WriterError(
                    ["Writer wrongly configured to not insert and to not update"]
                )

        except ValidationError as err:
            raise WriterError([{"ValidationError": err.messages}])
        except InvalidRelationValue as err:
            # TODO: Check if we can get the error message easier
            raise WriterError([{"InvalidRelationValue": err.args[0]}])

    def write_many(self, stream_entries, *args, **kwargs):
        """Writes the input entries using a given service."""
        entries = [entry.entry for entry in stream_entries]
        entries_with_id = [(self._entry_id(entry), entry) for entry in entries]
        results = self._service.create_or_update_many(self._identity, entries_with_id)
        stream_entries_processed = []
        for entry, result in zip(entries, results):
            processed_stream_entry = StreamEntry(
                entry=entry,
                record=result.record,
                errors=result.errors,
                op_type=result.op_type,
                exc=result.exc,
            )
            processed_stream_entry.log_errors()
            stream_entries_processed.append(processed_stream_entry)

        return stream_entries_processed


class YamlWriter(BaseWriter):
    """Writes the entries to a YAML file."""

    def __init__(self, filepath, *args, **kwargs):
        """Constructor.

        :param filepath: path of the output file.
        """
        self._filepath = Path(filepath)

        super().__init__(*args, **kwargs)

    def write(self, stream_entry, *args, **kwargs):
        """Writes the input stream entry using a given service."""
        with open(self._filepath, "a") as file:
            # made into array for safer append
            # will always read array (good for reader)
            yaml.safe_dump([stream_entry.entry], file, allow_unicode=True)

        return stream_entry

    def write_many(self, stream_entries, *args, **kwargs):
        """Writes the yaml input entries."""
        with open(self._filepath, "a") as file:
            yaml.safe_dump(
                [stream_entry.entry for stream_entry in stream_entries],
                file,
                allow_unicode=True,
            )


class AsyncWriter(BaseWriter):
    """Writes the entries asynchronously (celery task)."""

    def __init__(self, writer, *args, **kwargs):
        """Constructor.

        :param writer: writer to use.
        """
        super().__init__(*args, **kwargs)
        self._writer = writer

    def write(self, stream_entry, *args, **kwargs):
        """Launches a celery task to write an entry."""
        write_entry.delay(self._writer, stream_entry.entry)

        return stream_entry

    def write_many(self, stream_entries, *args, **kwargs):
        """Launches a celery task to write an entry."""
        write_many_entry.delay(
            self._writer, [stream_entry.entry for stream_entry in stream_entries]
        )

        return stream_entries



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022-2024 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Localization serializer for Funders."""

from marshmallow import Schema, fields

from invenio_vocabularies.resources import L10NString


class IdentifierSchema(Schema):
    """Identifier scheme."""

    identifier = fields.String(dump_only=True)
    scheme = fields.String(dump_only=True)


class FunderL10NItemSchema(Schema):
    """Funder serializer schema."""

    id = fields.String(dump_only=True)
    title = L10NString(data_key="title_l10n")
    description = L10NString(data_key="description_l10n")
    props = fields.Dict(dump_only=True)
    name = fields.String(dump_only=True)
    country = fields.String(dump_only=True)
    country_name = fields.String(dump_only=True)
    identifiers = fields.List(fields.Nested(IdentifierSchema), dump_only=True)



#
# Copyright (C) 2021 Northwestern University.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Create vocabulary schemes."""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "4a9a4fd235f8"
down_revision = "6312f33645c1"
branch_labels = ()
depends_on = None


def upgrade():
    """Upgrade database."""
    op.create_table(
        "vocabularies_schemes",
        sa.Column("id", sa.String(255), nullable=False),
        sa.Column("parent_id", sa.String(255), nullable=False),
        sa.Column("name", sa.String(255), nullable=True),
        sa.Column("uri", sa.String(255), nullable=True),
        sa.PrimaryKeyConstraint(
            "id", "parent_id", name=op.f("pk_vocabularies_schemes")
        ),
    )


def downgrade():
    """Downgrade database."""
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("vocabularies_schemes")
    # ### end Alembic commands ###



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2022 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Test the funder vocabulary service."""

import arrow
import pytest
from invenio_pidstore.errors import PIDAlreadyExists, PIDDeletedError
from marshmallow.exceptions import ValidationError

from invenio_vocabularies.contrib.funders.api import Funder


def test_simple_flow(app, service, identity, funder_full_data):
    """Test a simple vocabulary service flow."""
    # Service
    assert service.id == "funders"
    assert service.config.indexer_queue_name == "funders"

    # Create it
    item = service.create(identity, funder_full_data)
    id_ = item.id

    assert item.id == funder_full_data["id"]
    for k, v in funder_full_data.items():
        assert item.data[k] == v

    # Read it
    read_item = service.read(identity, "01ggx4157")

    assert item.id == read_item.id
    assert item.data == read_item.data

    # Refresh index to make changes live.
    Funder.index.refresh()

    # Search it
    res = service.search(identity, q=f"id:{id_}", size=25, page=1)
    assert res.total == 1
    assert list(res.hits)[0] == read_item.data

    # Update its country
    data = read_item.data
    data["country"] = "New country"
    update_item = service.update(identity, id_, data)
    assert item.id == update_item.id
    assert update_item["country"] == "New country"

    # Update its title
    data = read_item.data
    data["title"]["en"] = "New title"
    update_item = service.update(identity, id_, data)
    # the pid is not save to the json metadata
    assert not update_item._record.get("pid")
    assert item.id == update_item.id
    assert update_item["title"]["en"] == "New title"

    # Delete it
    assert service.delete(identity, id_)

    # Refresh to make changes live
    Funder.index.refresh()

    # Fail to retrieve it
    # - db
    # only the metadata is removed from the record, it is still resolvable
    base_keys = {"created", "updated", "id", "links", "revision_id"}
    deleted_rec = service.read(identity, id_).to_dict()
    assert set(deleted_rec.keys()) == base_keys
    # - search
    res = service.search(identity, q=f"id:{id_}", size=25, page=1)
    assert res.total == 0

    # not-ideal cleanup
    item._record.delete(force=True)


def test_pid_already_registered(
    app, service, identity, funder_full_data, example_funder
):
    """Recreating a record with same id should fail."""
    # example_funder does the first creation
    pytest.raises(PIDAlreadyExists, service.create, identity, funder_full_data)


def test_extra_fields(app, service, identity, funder_full_data):
    """Extra fields in data should fail."""
    funder_full_data["invalid"] = 1
    pytest.raises(ValidationError, service.create, identity, funder_full_data)


def test_indexed_at_query(app, db, service, identity, funder_full_data):
    before = arrow.utcnow().strftime("%Y-%m-%dT%H:%M:%S.%f")
    _ = service.create(identity, funder_full_data)
    now = arrow.utcnow().strftime("%Y-%m-%dT%H:%M:%S.%f")
    Funder.index.refresh()

    # there is previous to before
    res = service.search(identity, q=f"indexed_at:[* TO {before}]", size=25, page=1)
    assert res.total == 0

    # there is previous to now
    res = service.search(identity, q=f"indexed_at:[* TO {now}]", size=25, page=1)
    assert res.total == 1



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Vocabulary names configuration."""

from flask import current_app
from invenio_i18n import lazy_gettext as _
from invenio_records_resources.services import SearchOptions
from invenio_records_resources.services.records.components import (
    DataComponent,
    RelationsComponent,
)
from invenio_records_resources.services.records.params import SuggestQueryParser
from werkzeug.local import LocalProxy

from ...services.components import PIDComponent

names_schemes = LocalProxy(lambda: current_app.config["VOCABULARIES_NAMES_SCHEMES"])


class NamesSearchOptions(SearchOptions):
    """Search options."""

    suggest_parser_cls = SuggestQueryParser.factory(
        fields=[
            "given_name^100",
            "name^70",
            "family_name^50",
            "identifiers.identifier^20",
            "affiliations.name^20",
        ],
        type="most_fields",  # https://www.elastic.co/guide/en/elasticsearch/reference/current/query-dsl-multi-match-query.html#multi-match-types
        fuzziness="AUTO",
    )

    sort_default = "bestmatch"

    sort_default_no_query = "name"

    sort_options = {
        "bestmatch": dict(
            title=_("Best match"),
            fields=["_score"],  # ES defaults to desc on `_score` field
        ),
        "name": dict(
            title=_("Name"),
            fields=["name_sort"],
        ),
        "newest": dict(
            title=_("Newest"),
            fields=["-created"],
        ),
        "oldest": dict(
            title=_("Oldest"),
            fields=["created"],
        ),
    }


service_components = [
    # Order of components are important!
    DataComponent,
    PIDComponent,
    RelationsComponent,
]



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021-2022 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Fixtures module."""

import yaml
from invenio_access.permissions import system_identity

from .datastreams.factories import DataStreamFactory
from .proxies import current_service


class VocabularyFixture:
    """Vocabulary fixture."""

    def __init__(self, filepath, delay=True):
        """Constructor."""
        self._filepath = filepath

    def _load_vocabulary(self, config, delay=True, **kwargs):
        """Given an entry from the vocabularies.yaml file, load its content."""
        datastream = DataStreamFactory.create(
            readers_config=config["readers"],
            transformers_config=config.get("transformers"),
            writers_config=config["writers"],
        )

        errors = []
        for result in datastream.process():
            if result.errors:
                errors.append(result)

        return errors

    def _create_vocabulary(self, id_, pid_type, *args, **kwargs):
        """Creates a vocabulary."""
        return current_service.create_type(system_identity, id_, pid_type)

    def load(self, *args, **kwargs):
        """Return content of vocabularies file."""
        with open(self._filepath) as f:
            data = yaml.safe_load(f) or {}
            for id_, config in data.items():
                self._create_vocabulary(id_, config["pid-type"])
                yield self._load_vocabulary(config)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Vocabularies test config."""

from functools import partial

import pytest
from invenio_db import db
from invenio_indexer.api import RecordIndexer
from invenio_search import current_search_client

from invenio_vocabularies.records.api import Vocabulary
from invenio_vocabularies.records.models import VocabularyType


@pytest.fixture()
def indexer():
    """Indexer instance with correct Record class."""
    return RecordIndexer(
        record_cls=Vocabulary,
        record_to_index=lambda r: (r.__class__.index._name, "_doc"),
    )


@pytest.fixture()
def search_get():
    """Get a document from an index."""
    return partial(current_search_client.get, Vocabulary.index._name)


@pytest.fixture()
def lic_type():
    """Get a language vocabulary type."""
    return VocabularyType.create(id="licenses", pid_type="lic")


@pytest.fixture()
def example_data():
    """Example data."""
    return {
        "id": "eng",
        "title": {"en": "English", "da": "Engelsk"},
        "description": {"en": "Text", "da": "Tekst"},
        "icon": "file-o",
        "props": {
            "datacite_type": "Text",
        },
    }


@pytest.fixture()
def example_record(db, example_data, lang_type):
    """Example record."""
    record = Vocabulary.create(example_data, type=lang_type)
    Vocabulary.pid.create(record)
    record.commit()
    db.session.commit()
    return record



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2024 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Vocabulary permissions."""

from invenio_records_permissions.generators import AuthenticatedUser, SystemProcess

from ...services.permissions import PermissionPolicy


class NamesPermissionPolicy(PermissionPolicy):
    """Permission policy."""

    can_search = [SystemProcess(), AuthenticatedUser()]
    can_read = [SystemProcess(), AuthenticatedUser()]



# -*- coding: utf-8 -*-
#
# Copyright (C) 2024 CERN.
# Copyright (C) 2024 University of Münster.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Vocabulary services configs."""

import sqlalchemy as sa
from flask import current_app
from invenio_i18n import lazy_gettext as _
from invenio_records_resources.services import (
    Link,
    LinksTemplate,
    RecordService,
    RecordServiceConfig,
    SearchOptions,
    pagination_links,
)
from invenio_records_resources.services.base import (
    ConditionalLink,
    Service,
    ServiceListResult,
)
from invenio_records_resources.services.records.components import DataComponent
from invenio_records_resources.services.records.params import (
    FilterParam,
    SuggestQueryParser,
)

from ..records.api import Vocabulary
from ..records.models import VocabularyType
from . import results
from .components import PIDComponent, VocabularyTypeComponent
from .permissions import PermissionPolicy
from .schema import TaskSchema, VocabularySchema


def is_custom_vocabulary_type(vocabulary_type, context):
    """Check if the vocabulary type is a custom vocabulary type."""
    return vocabulary_type["id"] in current_app.config.get(
        "VOCABULARIES_CUSTOM_VOCABULARY_TYPES", []
    )


class VocabularySearchOptions(SearchOptions):
    """Search options for vocabularies."""

    params_interpreters_cls = [
        FilterParam.factory(param="tags", field="tags"),
    ] + SearchOptions.params_interpreters_cls

    suggest_parser_cls = SuggestQueryParser.factory(
        fields=[
            "id.text^100",
            "id.text._2gram",
            "id.text._3gram",
            "title.en^5",
            "title.en._2gram",
            "title.en._3gram",
        ],
    )

    sort_default = "bestmatch"

    sort_default_no_query = "title"

    sort_options = {
        "bestmatch": dict(
            title=_("Best match"),
            fields=["_score"],  # ES defaults to desc on `_score` field
        ),
        "title": dict(
            title=_("Title"),
            fields=["title_sort"],
        ),
        "newest": dict(
            title=_("Newest"),
            fields=["-created"],
        ),
        "oldest": dict(
            title=_("Oldest"),
            fields=["created"],
        ),
    }


class VocabularyTypeSearchOptions(SearchOptions):
    """Search options for vocabulary types."""

    sort_options = {
        "id": dict(
            title=_("ID"),
            fields=["id"],
        ),
    }

    sort_default = "id"

    sort_default_no_query = "id"

    sort_direction_options = {
        "asc": dict(title=_("Ascending"), fn=sa.asc),
        "desc": dict(title=_("Descending"), fn=sa.desc),
    }

    sort_direction_default = "asc"


class VocabulariesServiceConfig(RecordServiceConfig):
    """Vocabulary service configuration."""

    service_id = "vocabularies"
    indexer_queue_name = "vocabularies"
    permission_policy_cls = PermissionPolicy
    record_cls = Vocabulary
    schema = VocabularySchema
    task_schema = TaskSchema

    search = VocabularySearchOptions

    components = [
        # Order of components are important!
        VocabularyTypeComponent,
        DataComponent,
        PIDComponent,
    ]

    links_item = {
        "self": Link(
            "{+api}/vocabularies/{type}/{id}",
            vars=lambda record, vars: vars.update(
                {
                    "id": record.pid.pid_value,
                    "type": record.type.id,
                }
            ),
        ),
    }

    links_search = pagination_links("{+api}/vocabularies/{type}{?args*}")


class VocabularyTypesServiceConfig(RecordServiceConfig):
    """Vocabulary types service configuration."""

    service_id = "vocabulary_types"
    permission_policy_cls = PermissionPolicy
    record_cls = VocabularyType
    schema = VocabularySchema  # Works but should be VocabularyTypeSchema if this is defined at some point
    result_list_cls = results.VocabularyTypeList

    links_item = {
        "self": ConditionalLink(
            cond=is_custom_vocabulary_type,
            if_=Link(
                "{+api}/{id}",
                vars=lambda vocab_type, vars: vars.update({"id": vocab_type["id"]}),
            ),
            else_=Link(
                "{+api}/vocabularies/{id}",
                vars=lambda vocab_type, vars: vars.update({"id": vocab_type["id"]}),
            ),
        )
    }

    search = VocabularyTypeSearchOptions

    components = [
        # Order of components are important!
        VocabularyTypeComponent,
        DataComponent,
        PIDComponent,
    ]

    links_search = pagination_links("{+api}/vocabularies/{type}{?args*}")



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022-2024 CERN.
# Copyright (C) 2024 California Institute of Technology.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Affiliations datastreams tests."""

from copy import deepcopy

import pytest
from invenio_access.permissions import system_identity

from invenio_vocabularies.contrib.affiliations.api import Affiliation
from invenio_vocabularies.contrib.affiliations.config import affiliation_schemes
from invenio_vocabularies.contrib.affiliations.datastreams import (
    AffiliationsServiceWriter,
    OpenAIREAffiliationsServiceWriter,
    OpenAIREOrganizationTransformer,
)
from invenio_vocabularies.contrib.common.ror.datastreams import RORTransformer
from invenio_vocabularies.datastreams import StreamEntry
from invenio_vocabularies.datastreams.errors import TransformerError, WriterError


@pytest.fixture(scope="module")
def expected_from_ror_json():
    return {
        "id": "05dxps055",
        "name": "California Institute of Technology",
        "title": {
            "en": "California Institute of Technology",
            "es": "Instituto de Tecnología de California",
        },
        "acronym": "CIT",
        "aliases": ["Caltech"],
        "country": "US",
        "country_name": "United States",
        "location_name": "Pasadena",
        "status": "active",
        "identifiers": [
            {"scheme": "ror", "identifier": "05dxps055"},
            {"scheme": "grid", "identifier": "grid.20861.3d"},
            {"scheme": "isni", "identifier": "0000 0001 0706 8890"},
        ],
        "types": ["education", "funder"],
    }


def test_ror_transformer(app, dict_ror_entry, expected_from_ror_json):
    """Test RORTransformer to ensure it transforms ROR entries correctly."""
    transformer = RORTransformer(vocab_schemes=affiliation_schemes)
    assert expected_from_ror_json == transformer.apply(dict_ror_entry).entry


def test_affiliations_service_writer_create(app, search_clear, affiliation_full_data):
    """Test AffiliationsServiceWriter for creating a new affiliation."""
    writer = AffiliationsServiceWriter()
    affiliation_rec = writer.write(StreamEntry(affiliation_full_data))
    affiliation_dict = affiliation_rec.entry.to_dict()
    assert dict(affiliation_dict, **affiliation_full_data) == affiliation_dict

    # not-ideal cleanup
    affiliation_rec.entry._record.delete(force=True)


def test_affiliations_service_writer_duplicate(
    app, search_clear, affiliation_full_data
):
    """Test AffiliationsServiceWriter for handling duplicate entries."""
    writer = AffiliationsServiceWriter()
    affiliation_rec = writer.write(stream_entry=StreamEntry(affiliation_full_data))
    Affiliation.index.refresh()  # refresh index to make changes live
    with pytest.raises(WriterError) as err:
        writer.write(stream_entry=StreamEntry(affiliation_full_data))

    expected_error = [f"Vocabulary entry already exists: {affiliation_full_data}"]
    assert expected_error in err.value.args

    # not-ideal cleanup
    affiliation_rec.entry._record.delete(force=True)


def test_affiliations_service_writer_update_existing(
    app, search_clear, affiliation_full_data, service
):
    """Test updating an existing affiliation using AffiliationsServiceWriter."""
    # create vocabulary
    writer = AffiliationsServiceWriter(update=True)
    orig_affiliation_rec = writer.write(stream_entry=StreamEntry(affiliation_full_data))
    Affiliation.index.refresh()  # refresh index to make changes live
    # update vocabulary
    updated_affiliation = deepcopy(affiliation_full_data)
    updated_affiliation["name"] = "Updated Name"

    # check changes vocabulary
    _ = writer.write(stream_entry=StreamEntry(updated_affiliation))
    affiliation_rec = service.read(system_identity, orig_affiliation_rec.entry.id)

    affiliation_dict = affiliation_rec.to_dict()

    # needed while the writer resolves from ES
    assert _.entry.id == orig_affiliation_rec.entry.id
    assert dict(affiliation_dict, **updated_affiliation) == affiliation_dict

    # not-ideal cleanup
    affiliation_rec._record.delete(force=True)


def test_affiliations_service_writer_update_non_existing(
    app, search_clear, affiliation_full_data, service
):
    """Test creating a new affiliation when updating a non-existing entry."""
    # vocabulary item not created, call update directly
    updated_affiliation = deepcopy(affiliation_full_data)
    updated_affiliation["name"] = "New name"
    # check changes vocabulary
    writer = AffiliationsServiceWriter(update=True)
    affiliation_rec = writer.write(stream_entry=StreamEntry(updated_affiliation))
    affiliation_rec = service.read(system_identity, affiliation_rec.entry.id)
    affiliation_dict = affiliation_rec.to_dict()

    assert dict(affiliation_dict, **updated_affiliation) == affiliation_dict

    # not-ideal cleanup
    affiliation_rec._record.delete(force=True)


@pytest.fixture()
def dict_openaire_organization_entry():
    """An example entry from OpenAIRE organization Data Dump."""
    return StreamEntry(
        {
            "alternativeNames": [
                "European Organization for Nuclear Research",
                "Organisation européenne pour la recherche nucléaire",
                "CERN",
            ],
            "country": {"code": "CH", "label": "Switzerland"},
            "id": "openorgs____::47efb6602225236c0b207761a8b3a21c",
            "legalName": "European Organization for Nuclear Research",
            "legalShortName": "CERN",
            "pid": [
                {"scheme": "mag_id", "value": "67311998"},
                {"scheme": "ISNI", "value": "000000012156142X"},
                {"scheme": "Wikidata", "value": "Q42944"},
                {"scheme": "ROR", "value": "https://ror.org/01ggx4157"},
                {"scheme": "PIC", "value": "999988133"},
                {"scheme": "OrgReg", "value": "INT1011"},
                {"scheme": "ISNI", "value": "000000012156142X"},
                {"scheme": "FundRef", "value": "100012470"},
                {"scheme": "GRID", "value": "grid.9132.9"},
                {"scheme": "OrgRef", "value": "37351"},
            ],
            "websiteUrl": "http://home.web.cern.ch/",
        }
    )


@pytest.fixture(scope="module")
def expected_from_openaire_json():
    return {
        "id": "01ggx4157",
        "identifiers": [{"identifier": "999988133", "scheme": "pic"}],
    }


def test_openaire_organization_transformer(
    app, dict_openaire_organization_entry, expected_from_openaire_json
):
    """Test OpenAIREOrganizationTransformer for transforming OpenAIRE entries."""
    transformer = OpenAIREOrganizationTransformer()
    assert (
        expected_from_openaire_json
        == transformer.apply(dict_openaire_organization_entry).entry
    )


def test_openaire_affiliations_service_writer(
    app,
    search_clear,
    affiliation_full_data,
    openaire_affiliation_full_data,
    service,
):
    """Test writing and updating an OpenAIRE affiliation entry."""
    # create vocabulary with original service writer
    orig_writer = AffiliationsServiceWriter()

    orig_affiliation_rec = orig_writer.write(StreamEntry(affiliation_full_data))

    orig_affiliation_dict = orig_affiliation_rec.entry.to_dict()
    Affiliation.index.refresh()  # refresh index to make changes live

    # update vocabulary and check changes vocabulary with OpenAIRE service writer
    writer = OpenAIREAffiliationsServiceWriter()

    _ = writer.write(StreamEntry(openaire_affiliation_full_data))
    Affiliation.index.refresh()  # refresh index to make changes live
    affiliation_rec = service.read(system_identity, orig_affiliation_rec.entry.id)
    affiliation_dict = affiliation_rec.to_dict()

    assert _.entry.id == orig_affiliation_rec.entry.id

    # updating fields changing from one update to the other
    orig_affiliation_dict["revision_id"] = affiliation_dict["revision_id"]
    orig_affiliation_dict["updated"] = affiliation_dict["updated"]
    # Adding the extra identifier coming from OpenAIRE
    orig_affiliation_dict["identifiers"].extend(
        openaire_affiliation_full_data["identifiers"]
    )

    assert dict(orig_affiliation_dict) == affiliation_dict

    # not-ideal cleanup
    affiliation_rec._record.delete(force=True)


def test_openaire_affiliations_transformer_non_openorgs(
    app, dict_openaire_organization_entry
):
    """Test error handling for non-OpenOrgs ID in OpenAIRE transformer."""
    transformer = OpenAIREOrganizationTransformer()

    updated_organization_entry = deepcopy(dict_openaire_organization_entry.entry)
    updated_organization_entry["id"] = "pending_org_::627931d047132a4e20dbc4a882eb9a35"

    with pytest.raises(TransformerError) as err:
        transformer.apply(StreamEntry(updated_organization_entry))

    expected_error = [
        f"Not valid OpenAIRE OpenOrgs id for: {updated_organization_entry}"
    ]
    assert expected_error in err.value.args


def test_openaire_affiliations_transformer_no_id(app, dict_openaire_organization_entry):
    """Test error handling when missing ID in OpenAIRE transformer."""
    transformer = OpenAIREOrganizationTransformer()

    updated_organization_entry = deepcopy(dict_openaire_organization_entry.entry)
    updated_organization_entry.pop("id", None)

    with pytest.raises(TransformerError) as err:
        transformer.apply(StreamEntry(updated_organization_entry))

    expected_error = [f"No id for: {updated_organization_entry}"]
    assert expected_error in err.value.args


def test_openaire_affiliations_transformer_no_alternative_identifiers(
    app, dict_openaire_organization_entry
):
    """Test error handling when missing alternative identifiers in OpenAIRE transformer."""
    transformer = OpenAIREOrganizationTransformer()

    updated_organization_entry = deepcopy(dict_openaire_organization_entry.entry)
    updated_organization_entry.pop("pid", None)

    with pytest.raises(TransformerError) as err:
        transformer.apply(StreamEntry(updated_organization_entry))

    expected_error = [f"No alternative identifiers for: {updated_organization_entry}"]
    assert expected_error in err.value.args



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Pytest configuration.

See https://pytest-invenio.readthedocs.io/ for documentation on which test
fixtures are available.
"""

import pytest
from invenio_i18n import lazy_gettext as _
from invenio_indexer.proxies import current_indexer_registry
from invenio_records_resources.proxies import current_service_registry

from invenio_vocabularies.contrib.awards.api import Award
from invenio_vocabularies.contrib.funders.api import Funder


#
# Funder related fixtures
#
@pytest.fixture(scope="function")
def example_funder(db, identity, funders_service, funder_indexer):
    """Example funder."""
    funder_data = {
        "id": "01ggx4157",
        "name": "CERN",
        "title": {
            "en": "European Organization for Nuclear Research",
            "fr": "Organisation européenne pour la recherche nucléaire",
        },
        "country": "CH",
    }
    funder = funders_service.create(identity, funder_data)
    Funder.index.refresh()  # Refresh the index
    yield funder
    funder._record.delete(force=True)
    funder_indexer.delete(funder._record, refresh=True)
    db.session.commit()


@pytest.fixture(scope="function")
def example_funder_ec(db, identity, funders_service, funder_indexer):
    """Example European Commission funder."""
    funder_data = {
        "id": "00k4n6c32",
        "name": "EC",
        "title": {"en": "European Commission", "fr": "Commission Européenne"},
        "country": "BE",
    }
    funder = funders_service.create(identity, funder_data)
    Funder.index.refresh()  # Refresh the index
    yield funder
    funder._record.delete(force=True)
    funder_indexer.delete(funder._record, refresh=True)
    db.session.commit()


@pytest.fixture(scope="module")
def funders_service():
    """Funders service object."""
    return current_service_registry.get("funders")


@pytest.fixture()
def funder_indexer():
    """Indexer instance with correct Record class."""
    return current_indexer_registry.get("funders")


#
# Award related fixtures
#
@pytest.fixture(scope="function")
def award_full_data():
    """Full award data."""
    return {
        "id": "755021",
        "identifiers": [
            {
                "identifier": "https://cordis.europa.eu/project/id/755021",
                "scheme": "url",
            }
        ],
        "number": "755021",
        "title": {
            "en": "Personalised Treatment For Cystic Fibrosis Patients With \
                Ultra-rare CFTR Mutations (and beyond)",
        },
        "funder": {"id": "00k4n6c32"},
        "acronym": "HIT-CF",
        "program": "H2020",
    }


@pytest.fixture(scope="function")
def award_full_data_invalid_id():
    """Full award data."""
    return {
        "id": "755021",
        "identifiers": [
            {
                "identifier": "https://cordis.europa.eu/project/id/755021",
                "scheme": "url",
            }
        ],
        "number": "755021",
        "title": {
            "en": "Personalised Treatment For Cystic Fibrosis Patients With \
                Ultra-rare CFTR Mutations (and beyond)",
        },
        "funder": {"id": "010101010"},
        "acronym": "HIT-CF",
        "program": "H2020",
    }


@pytest.fixture(scope="function")
def example_award(db, example_funder_ec, award_full_data, identity, indexer, service):
    """Creates and hard deletes an award."""
    award = service.create(identity, award_full_data)
    Award.index.refresh()  # Refresh the index
    yield award
    award._record.delete(force=True)
    indexer.delete(award._record, refresh=True)
    db.session.commit()


@pytest.fixture(scope="module")
def service():
    """Awards service object."""
    return current_service_registry.get("awards")


@pytest.fixture()
def indexer():
    """Indexer instance with correct Record class."""
    return current_indexer_registry.get("awards")



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022-2024 CERN.
# Copyright (C) 2024 California Institute of Technology.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""MeSH subjects datastreams, transformers, writers and readers."""

from invenio_vocabularies.datastreams.transformers import (
    BaseTransformer,
    TransformerError,
)


class MeshSubjectsTransformer(BaseTransformer):
    """MeSH subjects Transformer."""

    def apply(self, stream_entry, *args, **kwargs):
        """Apply transformation on steam entry."""
        entry_data = stream_entry.entry

        # ID in MeSH data is the URL, ex. https://id.nlm.nih.gov/mesh/D000001
        # We just want to use the ID prefixed by "mesh:""
        try:
            mesh_id = entry_data["id"].split("/")[-1]
        except Exception:
            raise TransformerError("Not a valid MeSH ID.")

        entry_data["id"] = "mesh:" + mesh_id
        return stream_entry


VOCABULARIES_DATASTREAM_READERS = {}
"""MeSH datastream readers."""

VOCABULARIES_DATASTREAM_WRITERS = {}
"""MeSH subject datastream writers."""

VOCABULARIES_DATASTREAM_TRANSFORMERS = {"mesh-subjects": MeshSubjectsTransformer}
"""MeSH subjects datastream transformers."""



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
# Copyright (C) 2021 Northwestern University.
# Copyright (C) 2024 University of Münster.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Subjects configuration."""

from flask import current_app
from invenio_i18n import lazy_gettext as _
from invenio_records_resources.services import SearchOptions
from invenio_records_resources.services.records.components import DataComponent
from werkzeug.local import LocalProxy

from ...services.components import PIDComponent
from ...services.querystr import FilteredSuggestQueryParser

subject_schemes = LocalProxy(
    lambda: current_app.config["VOCABULARIES_SUBJECTS_SCHEMES"]
)


class SubjectsSearchOptions(SearchOptions):
    """Search options."""

    suggest_parser_cls = FilteredSuggestQueryParser.factory(
        filter_field="scheme",
        fields=[  # suggest fields
            "title.*^100",
            "synonyms^20",
        ],
    )

    sort_default = "bestmatch"

    sort_default_no_query = "subject"

    sort_options = {
        "bestmatch": dict(
            title=_("Best match"),
            fields=["_score"],  # ES defaults to desc on `_score` field
        ),
        "subject": dict(
            title=_("Name"),
            fields=["subject_sort"],
        ),
        "newest": dict(
            title=_("Newest"),
            fields=["-created"],
        ),
        "oldest": dict(
            title=_("Oldest"),
            fields=["created"],
        ),
    }


service_components = [
    # Order of components are important!
    DataComponent,
    PIDComponent,
]



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021-2024 CERN.
# Copyright (C)      2024 University of Münster.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Readers module."""

import csv
import gzip
import json
import re
import tarfile
import zipfile
from abc import ABC, abstractmethod
from json.decoder import JSONDecodeError

import requests
import yaml
from lxml import etree
from lxml.html import fromstring
from lxml.html import parse as html_parse

from .errors import ReaderError
from .xml import etree_to_dict

try:
    import oaipmh_scythe
except ImportError:
    oaipmh_scythe = None


class BaseReader(ABC):
    """Base reader."""

    def __init__(self, origin=None, mode="r", *args, **kwargs):
        """Constructor.

        :param origin: Data source (e.g. filepath).
                       Can be none in case of piped readers.
        """
        self._origin = origin
        self._mode = mode

    @abstractmethod
    def _iter(self, fp, *args, **kwargs):
        """Yields data objects file pointer."""
        pass

    def read(self, item=None, *args, **kwargs):
        """Reads from item or opens the file descriptor from origin."""
        if item:
            yield from self._iter(fp=item, *args, **kwargs)
        else:
            with open(self._origin, self._mode) as file:
                yield from self._iter(fp=file, *args, **kwargs)


class YamlReader(BaseReader):
    """Yaml reader."""

    def _iter(self, fp, *args, **kwargs):
        """Reads a yaml file and returns a dictionary per element."""
        data = yaml.safe_load(fp) or []
        for entry in data:
            yield entry


class TarReader(BaseReader):
    """Tar reader."""

    def __init__(self, *args, mode="r|gz", regex=None, **kwargs):
        """Constructor."""
        self._regex = re.compile(regex) if regex else None
        super().__init__(*args, mode=mode, **kwargs)

    def _iter(self, fp, *args, **kwargs):
        """Iterates through the files in the archive."""
        for member in fp:
            match = not self._regex or self._regex.search(member.name)
            if member.isfile() and match:
                yield fp.extractfile(member)

    def read(self, item=None, *args, **kwargs):
        """Opens a tar archive or uses the given file pointer."""
        if item:
            if isinstance(item, tarfile.TarFile):
                yield from self._iter(fp=item, *args, **kwargs)
            else:
                # If the item is not already a TarFile (e.g. if it is a BytesIO), try to create a TarFile from the item.
                with tarfile.open(mode=self._mode, fileobj=item) as archive:
                    yield from self._iter(fp=archive, *args, **kwargs)
        else:
            with tarfile.open(self._origin, self._mode) as archive:
                yield from self._iter(fp=archive, *args, **kwargs)


class SimpleHTTPReader(BaseReader):
    """Simple HTTP Reader."""

    def __init__(self, origin, id=None, ids=None, content_type=None, *args, **kwargs):
        """Constructor."""
        assert id or ids
        self._ids = ids if ids else [id]
        self.content_type = content_type
        super().__init__(origin, *args, **kwargs)

    def _iter(self, url, *args, **kwargs):
        """Queries an URL."""
        base_url = url
        headers = {"Accept": self.content_type}

        for id_ in self._ids:
            url = base_url.format(id=id_)
            resp = requests.get(url, headers=headers)
            if resp.status_code != 200:
                # todo add logging/fail
                pass

            yield resp.content

    def read(self, item=None, *args, **kwargs):
        """Chooses between item and origin as url."""
        url = item if item else self._origin
        yield from self._iter(url=url, *args, **kwargs)


class ZipReader(BaseReader):
    """ZIP reader."""

    def __init__(self, *args, options=None, regex=None, **kwargs):
        """Constructor."""
        self._options = options or {}
        self._regex = re.compile(regex) if regex else None
        super().__init__(*args, **kwargs)

    def _iter(self, fp, *args, **kwargs):
        """Iterates through the files in the archive."""
        for member in fp.infolist():
            match = not self._regex or self._regex.search(member.filename)
            if not member.is_dir() and match:
                yield fp.open(member)

    def read(self, item=None, *args, **kwargs):
        """Opens a Zip archive or uses the given file pointer."""
        # https://docs.python.org/3/library/zipfile.html
        if item:
            if isinstance(item, zipfile.ZipFile):
                yield from self._iter(fp=item, *args, **kwargs)
            else:
                # If the item is not already a ZipFile (e.g. if it is a BytesIO), try to create a ZipFile from the item.
                with zipfile.ZipFile(item, **self._options) as archive:
                    yield from self._iter(fp=archive, *args, **kwargs)
        else:
            with zipfile.ZipFile(self._origin, **self._options) as archive:
                yield from self._iter(fp=archive, *args, **kwargs)


class JsonReader(BaseReader):
    """JSON object reader."""

    def _iter(self, fp, *args, **kwargs):
        """Reads (loads) a json object and yields its items."""
        try:
            entries = json.load(fp)
            if isinstance(entries, list):
                for entry in entries:
                    yield entry
            else:
                yield entries  # just one entry
        except JSONDecodeError as err:
            raise ReaderError(f"Cannot decode JSON file {fp.name}: {str(err)}")


class JsonLinesReader(BaseReader):
    """JSON Lines reader."""

    def _iter(self, fp, *args, **kwargs):
        for idx, line in enumerate(fp):
            try:
                data = json.loads(line)
                if isinstance(data, list):
                    for entry in data:
                        yield entry
                else:
                    yield data  # just one entry
            except JSONDecodeError as err:
                raise ReaderError(
                    f"Cannot decode JSON line {fp.name}:{idx}: {str(err)}"
                )


class GzipReader(BaseReader):
    """Gzip reader."""

    def _iter(self, fp, *args, **kwargs):
        with gzip.open(fp) as gp:
            yield gp


class CSVReader(BaseReader):
    """Reads a CSV file and returns a dictionary per element."""

    def __init__(self, *args, csv_options=None, as_dict=True, **kwargs):
        """Constructor."""
        self.csv_options = csv_options or {}
        self.as_dict = as_dict
        super().__init__(*args, **kwargs)

    def _iter(self, fp, *args, **kwargs):
        """Reads a csv file and returns a dictionary per element."""
        csvfile = fp
        if self.as_dict:
            reader = csv.DictReader(csvfile, **self.csv_options)
        else:
            reader = csv.reader(csvfile, **self.csv_options)
        for row in reader:
            yield row


class XMLReader(BaseReader):
    """XML reader."""

    def __init__(self, root_element=None, *args, **kwargs):
        """Constructor."""
        self.root_element = root_element
        super().__init__(*args, **kwargs)

    def _iter(self, fp, *args, **kwargs):
        """Read and parse an XML file to dict."""
        # NOTE: We parse HTML, to skip XML validation and strip XML namespaces
        record = None
        try:
            xml_tree = fromstring(fp)
            xml_dict = etree_to_dict(xml_tree)
        except Exception as e:
            xml_tree = html_parse(fp).getroot()
            xml_dict = etree_to_dict(xml_tree)["html"]["body"]

        if self.root_element:
            record = xml_dict.get(self.root_element)
            if not record:
                raise ReaderError(
                    f"Root element '{self.root_element}' not found in XML entry."
                )
        else:
            record = xml_dict

        yield record


class OAIPMHReader(BaseReader):
    """OAIPMH reader."""

    def __init__(
        self,
        *args,
        base_url=None,
        metadata_prefix=None,
        set=None,
        from_date=None,
        until_date=None,
        verb=None,
        **kwargs,
    ):
        """Constructor."""
        self._base_url = base_url
        self._metadata_prefix = metadata_prefix if not None else "oai_dc"
        self._set = set
        self._until = until_date
        self._from = from_date
        self._verb = verb if not None else "ListRecords"
        super().__init__(*args, **kwargs)

    def _iter(self, scythe, *args, **kwargs):
        """Read and parse an OAIPMH stream to dict."""

        class OAIRecord(oaipmh_scythe.models.Record):
            """An XML unpacking implementation for more complicated formats."""

            def get_metadata(self):
                """Extract and return the record's metadata as a dictionary."""
                return xml_to_dict(
                    self.xml.find(f".//{self._oai_namespace}metadata").getchildren()[0],
                )

        if self._verb == "ListRecords":
            scythe.class_mapping["ListRecords"] = OAIRecord
            try:
                records = scythe.list_records(
                    from_=self._from,
                    until=self._until,
                    metadata_prefix=self._metadata_prefix,
                    set_=self._set,
                    ignore_deleted=True,
                )
                for record in records:
                    yield {"record": record}
            except oaipmh_scythe.NoRecordsMatch:
                raise ReaderError("No records found in OAI-PMH request.")
        else:
            scythe.class_mapping["GetRecord"] = OAIRecord
            try:
                headers = scythe.list_identifiers(
                    from_=self._from,
                    until=self._until,
                    metadata_prefix=self._metadata_prefix,
                    set_=self._set,
                    ignore_deleted=True,
                )
                for header in headers:
                    record = scythe.get_record(
                        identifier=header.identifier,
                        metadata_prefix=self._metadata_prefix,
                    )
                    yield {"record": record}
            except oaipmh_scythe.NoRecordsMatch:
                raise ReaderError("No records found in OAI-PMH request.")

    def read(self, item=None, *args, **kwargs):
        """Reads from item or opens the file descriptor from origin."""
        if item:
            raise NotImplementedError(
                "OAIPMHReader does not support being chained after another reader"
            )
        else:
            with oaipmh_scythe.Scythe(self._base_url) as scythe:
                yield from self._iter(scythe=scythe, *args, **kwargs)


def xml_to_dict(tree: etree._Element):
    """Convert an XML tree to a dictionary.

    This function takes an XML element tree and converts it into a dictionary.

    Args:
        tree: The root element of the XML tree to be converted.

    Returns:
        A dictionary with the key "record".
    """
    dict_obj = dict()
    dict_obj["record"] = etree.tostring(tree)

    return dict_obj



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 Northwestern University.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Test the subjects facets."""

from invenio_vocabularies.contrib.subjects import SubjectsLabels


def test_subjects_label_duplicates():
    labels_instance = SubjectsLabels()

    result = labels_instance(["MeSH", "MeSH"])

    assert {"MeSH": "MeSH"} == result

    result = labels_instance(["Abdomen", "Abdomen", "Abdomen"])

    assert {"Abdomen": "Abdomen"} == result


def test_subjects_label_empty():
    labels_instance = SubjectsLabels()

    result = labels_instance([])

    assert {} == result



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2024 CERN.
# Copyright (C) 2023 Graz University of Technology.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Resources layer tests."""

import pytest
from flask import g

from invenio_vocabularies.records.api import Vocabulary


#
# Fixtures
#
@pytest.fixture(scope="module")
def example_data():
    """Example data for records."""
    return {
        "id": "text",
        "title": {
            "en": "Text",
            "da": "Tekst",
        },
        "description": {
            "en": "Publications",
            "da": "Publikationer",
        },
        "icon": "file-o",
        "props": {},
        "type": "resourcetypes2",
    }


@pytest.fixture(scope="module")
def example_record(database, identity, service, example_data):
    """Create a vocabulary record."""
    service.create_type(identity, "resourcetypes2", "rt2")
    item = service.create(identity, example_data)
    Vocabulary.index.refresh()
    return item


@pytest.fixture()
def prefix():
    """API prefix."""
    return "/vocabularies/resourcetypes2"


@pytest.fixture()
def h():
    """Header for localised versions."""
    return {"accept": "application/vnd.inveniordm.v1+json"}


@pytest.fixture(scope="module")
def expected_da():
    """Expected serialization when danish chosen."""
    return {
        "id": "text",
        "title_l10n": "Tekst",
        "description_l10n": "Publikationer",
        "icon": "file-o",
        "props": {},
    }


@pytest.fixture(scope="module")
def expected_en():
    """Expected serialization when english chosen."""
    return {
        "id": "text",
        "title_l10n": "Text",
        "description_l10n": "Publications",
        "icon": "file-o",
        "props": {},
    }


#
# Tests
#
def test_get(client, example_record, h, prefix, expected_da, expected_en):
    """Test the endpoint to retrieve a single item."""
    id_ = example_record.id

    # Default locale
    res = client.get(f"{prefix}/{id_}", headers=h)
    assert res.json == expected_en

    # the used context is not reseted, every new client.get call uses the cache
    # which is bad
    g._flask_babel.babel_locale = None

    # Choose via querystring (?ln=da)
    res = client.get(f"{prefix}/{id_}?ln=da", headers=h)
    assert res.json == expected_da

    g._flask_babel.babel_locale = None

    res = client.get(f"{prefix}/{id_}?ln=en", headers=h)
    assert res.json == expected_en

    g._flask_babel.babel_locale = None

    # Choose via header
    h["accept-language"] = "da"
    res = client.get(f"{prefix}/{id_}", headers=h)
    assert res.json == expected_da

    g._flask_babel.babel_locale = None

    h["accept-language"] = "en"
    res = client.get(f"{prefix}/{id_}", headers=h)
    assert res.json == expected_en


def test_search(client, example_record, h, prefix, expected_da, expected_en):
    """Test search result serialization."""
    expected_en = {
        "hits": {"hits": [expected_en], "total": 1},
        "links": {
            "self": "https://127.0.0.1:5000/api/vocabularies/resourcetypes2?page=1&size=25&sort=title"
        },
        "sortBy": "title",
    }
    expected_da = {
        "hits": {"hits": [expected_da], "total": 1},
        "links": {
            "self": "https://127.0.0.1:5000/api/vocabularies/resourcetypes2?page=1&size=25&sort=title"
        },
        "sortBy": "title",
    }

    # Default locale
    res = client.get(f"{prefix}", headers=h)
    assert res.json == expected_en

    g._flask_babel.babel_locale = None

    # Choose via querystring (?ln=da)
    res = client.get(f"{prefix}?ln=da", headers=h)
    assert res.json == expected_da

    g._flask_babel.babel_locale = None

    res = client.get(f"{prefix}?ln=en", headers=h)
    assert res.json == expected_en

    g._flask_babel.babel_locale = None

    # Choose via header
    h["accept-language"] = "da"
    res = client.get(f"{prefix}", headers=h)
    assert res.json == expected_da

    g._flask_babel.babel_locale = None

    h["accept-language"] = "en"
    res = client.get(f"{prefix}", headers=h)
    assert res.json == expected_en



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Data access layer tests."""

import pytest
from jsonschema import ValidationError as SchemaValidationError
from sqlalchemy import inspect

from invenio_vocabularies.records.api import Vocabulary


def test_record_schema_validation(app, db, lang_type):
    """Record schema validation."""
    # Good data
    record = Vocabulary.create(
        {
            "id": "eng",
            "title": {"en": "English", "da": "Engelsk"},
            "description": {"en": "English", "da": "Engelsk"},
            "icon": "en",
            "props": {"akey": "a value"},
            "tags": ["recommended"],
        },
        type=lang_type,
    )
    assert record.schema

    # Bad data
    examples = [
        # title/descriptions are objects of key/string.
        {"id": "en", "title": "not a string"},
        {"id": "en", "title": {"en": 123}},
        {"id": "en", "description": "not a string"},
        # icon must be strings
        {"id": "en", "icon": 123},
        # props values must be strings
        {"id": "en", "props": {"key": 123}},
        {"id": "en", "props": {"key": {"test": "test"}}},
        # Additional properties false
        {"id": "en", "metadata": {"title": "test"}},
    ]

    for ex in examples:
        pytest.raises(SchemaValidationError, Vocabulary.create, ex)


def test_record_indexing(app, db, search, example_record, indexer, search_get):
    """Test indexing of a record."""
    # Index document in ES
    assert indexer.index(example_record)["result"] == "created"

    # Retrieve document from ES
    data = search_get(id=example_record.id)

    # Loads the ES data and compare
    record = Vocabulary.loads(data["_source"])
    assert record == example_record
    assert record.id == example_record.id
    assert record.revision_id == example_record.revision_id
    assert record.created == example_record.created
    assert record.updated == example_record.updated

    # Check system fields - i.e reading related type object from
    assert record == example_record
    assert record.type.id == "languages"
    assert record.type.pid_type == "lng"

    # Check that object was recrated without hitting DB
    assert inspect(record.type).persistent is False
    Vocabulary.type.session_merge(record)
    assert inspect(record.type).persistent is True


def test_record_pids(app, db, lang_type, lic_type):
    """Test record pid creation."""
    record = Vocabulary.create(
        {"id": "eng", "title": {"en": "English", "da": "Engelsk"}}, type=lang_type
    )
    Vocabulary.pid.create(record)
    assert record.type == lang_type
    assert record.pid.pid_value == "eng"
    assert record.pid.pid_type == "lng"
    assert Vocabulary.pid.resolve(("languages", "eng"))

    record = Vocabulary.create(
        {"id": "cc-by", "title": {"en": "CC-BY", "da": "CC-BY"}}, type=lic_type
    )
    Vocabulary.pid.create(record)
    assert record.type == lic_type
    assert record.pid.pid_value == "cc-by"
    assert record.pid.pid_type == "lic"
    assert Vocabulary.pid.resolve(("licenses", "cc-by"))



# -*- coding: utf-8 -*-
#
# Copyright (C) 2024 University of Münster.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Subject datastream tests."""

from copy import deepcopy

import pytest
from invenio_access.permissions import system_identity
from invenio_records_resources.proxies import current_service_registry

from invenio_vocabularies.contrib.subjects.api import Subject
from invenio_vocabularies.contrib.subjects.datastreams import SubjectsServiceWriter
from invenio_vocabularies.datastreams import StreamEntry
from invenio_vocabularies.datastreams.errors import WriterError


@pytest.fixture(scope="module")
def dict_subject_entry():
    return StreamEntry(
        {
            "title": {
                "en": "Dark Web",
                "de": "Darknet",
                "fr": "Réseaux anonymes (informatique)",
            },
            "id": "http://d-nb.info/gnd/1062531973",
            "scheme": "GND",
            "synonyms": ["Deep Web"],
        },
    )


def test_subjects_service_writer_create(app, db, search_clear, dict_subject_entry):
    writer = SubjectsServiceWriter()
    record = writer.write(dict_subject_entry)
    record = record.entry.to_dict()

    assert dict(record, **dict_subject_entry.entry) == record


def test_subjects_service_writer_duplicate(app, db, search_clear, dict_subject_entry):
    writer = SubjectsServiceWriter()
    _ = writer.write(stream_entry=dict_subject_entry)
    Subject.index.refresh()  # refresh index to make changes live
    with pytest.raises(WriterError) as err:
        writer.write(stream_entry=dict_subject_entry)

    expected_error = [f"Vocabulary entry already exists: {dict_subject_entry.entry}"]
    assert expected_error in err.value.args


def test_subjects_service_writer_update_existing(
    app, db, search_clear, dict_subject_entry
):
    # create vocabulary
    writer = SubjectsServiceWriter(update=True)
    subject = writer.write(stream_entry=dict_subject_entry)
    Subject.index.refresh()  # refresh index to make changes live
    # update vocabulary
    updated_subject = deepcopy(dict_subject_entry.entry)
    updated_subject["scheme"] = "new_scheme"
    del updated_subject["synonyms"]
    # check changes vocabulary
    _ = writer.write(stream_entry=StreamEntry(updated_subject))
    service = current_service_registry.get("subjects")
    record = service.read(system_identity, subject.entry.id)
    record = record.to_dict()

    # needed while the writer resolves from ES
    assert _.entry.id == subject.entry.id
    assert dict(record, **updated_subject) == record


def test_subjects_service_writer_update_non_existing(
    app, db, search_clear, dict_subject_entry
):
    # vocabulary item not created, call update directly
    updated_subject = deepcopy(dict_subject_entry.entry)
    updated_subject["scheme"] = "new_scheme"
    del updated_subject["synonyms"]
    # check changes vocabulary
    writer = SubjectsServiceWriter(update=True)
    subject = writer.write(stream_entry=StreamEntry(updated_subject))
    service = current_service_registry.get("subjects")
    record = service.read(system_identity, subject.entry.id)
    record = record.to_dict()

    assert dict(record, **updated_subject) == record



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2021 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Test the names vocabulary service."""

from copy import deepcopy

import arrow
import pytest
from invenio_pidstore.errors import PIDDoesNotExistError
from marshmallow.exceptions import ValidationError

from invenio_vocabularies.contrib.names.api import Name


def test_simple_flow(app, service, identity, name_full_data, example_affiliation):
    """Test a simple vocabulary service flow."""
    # Service
    assert service.id == "names"
    assert service.config.indexer_queue_name == "names"

    # Create it
    item = service.create(identity, name_full_data)
    id_ = item.id

    # add dereferenced values
    expected_data = deepcopy(name_full_data)
    expected_data["affiliations"][0][
        "name"
    ] = "European Organization for Nuclear Research"

    for k, v in expected_data.items():
        assert item.data[k] == v

    # Read it
    read_item = service.read(identity, id_)
    assert item.id == read_item.id
    assert item.data == read_item.data

    # Refresh index to make changes live.
    Name.index.refresh()

    # Search it
    res = service.search(identity, q=f"id:{id_}", size=25, page=1)
    assert res.total == 1
    assert list(res.hits)[0] == read_item.data

    # Update it
    data = read_item.data
    data["given_name"] = "Jane"
    update_item = service.update(identity, id_, data)
    assert item.id == update_item.id
    assert update_item["given_name"] == "Jane"
    assert update_item["name"] == "Doe, Jane"  # automatic update

    # Delete it
    assert service.delete(identity, id_)

    # Refresh to make changes live
    Name.index.refresh()

    # Fail to retrieve it
    # - db
    # only the metadata is removed from the record, it is still resolvable
    base_keys = {"created", "updated", "id", "links", "revision_id"}
    deleted_rec = service.read(identity, id_).to_dict()
    assert set(deleted_rec.keys()) == base_keys
    # - search
    res = service.search(identity, q=f"id:{id_}", size=25, page=1)
    assert res.total == 0


def test_extra_fields(app, service, identity, name_full_data):
    """Extra fields in data should fail."""
    name_full_data["invalid"] = 1
    pytest.raises(ValidationError, service.create, identity, name_full_data)


def test_identifier_resolution(
    app, search_clear, service, identity, name_full_data, example_affiliation
):
    # Create it
    item = service.create(identity, name_full_data)
    id_ = item.id

    Name.index.refresh()
    resolved = service.resolve(identity, id_="0000-0001-8135-3489", id_type="orcid")
    assert resolved.id == id_

    # non-existent orcid
    pytest.raises(
        PIDDoesNotExistError, service.resolve, identity, "0000-0002-5082-6404", "orcid"
    )

    # non-existent scheme
    pytest.raises(
        PIDDoesNotExistError,
        service.resolve,
        identity,
        "0000-0001-8135-3489",
        "invalid",
    )


def test_names_dereferenced(app, search_clear, service, identity, example_affiliation):
    """Extra fields in data should fail."""
    expected_aff = {"id": "cern", "name": "European Organization for Nuclear Research"}

    name_data = {
        "id": "0000-0001-8135-3489",
        "name": "Doe, John",
        "affiliations": [{"id": "cern"}],
    }

    # data is not dereferenced
    assert not name_data["affiliations"][0].get("name")
    # create it
    item = service.create(identity, name_data)
    Name.index.refresh()
    id_ = item.id
    assert item["affiliations"][0] == expected_aff

    # read it
    read_item = service.read(identity, id_)
    assert read_item["affiliations"][0] == expected_aff

    # search it
    res = service.search(identity, q=f"id:{id_}", size=25, page=1)
    assert res.total == 1
    assert list(res.hits)[0]["affiliations"][0] == expected_aff


def test_indexed_at_query(
    app, db, service, identity, name_full_data, example_affiliation
):
    before = arrow.utcnow().strftime("%Y-%m-%dT%H:%M:%S.%f")
    _ = service.create(identity, name_full_data)
    now = arrow.utcnow().strftime("%Y-%m-%dT%H:%M:%S.%f")
    Name.index.refresh()

    # there is previous to before
    res = service.search(identity, q=f"indexed_at:[* TO {before}]", size=25, page=1)
    assert res.total == 0

    # there is previous to now
    res = service.search(identity, q=f"indexed_at:[* TO {now}]", size=25, page=1)
    assert res.total == 1



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Vocabulary awards facets and labels."""

from ...services.facets import VocabularyLabels, lazy_get_label


class FundersLabels(VocabularyLabels):
    """Fetching of vocabulary labels for facets."""

    def __init__(self, vocabulary, cache=True, cache_ttl=3600, service_id=None):
        """Initialize the labels.

        :param vocabulary: the name of the vocabulary type.
        :param cache: use simple process in-memory cache when True.
        :param cache_ttl: cache expiration in seconds.
        :param service_id: the id of the registered service to be used
            when fetching values for the vocabulary.
        """
        super().__init__(
            vocabulary,
            cache=cache,
            cache_ttl=cache_ttl,
            service_id="funders",
        )
        self.fields = ("id", "title", "country")  # not configurable

    def _vocab_to_label(self, vocab):
        """Returns the label string for a vocabulary entry."""
        return f"{lazy_get_label(vocab['title'])} ({vocab['country']})"



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2024 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""S3 client."""

from flask import current_app

try:
    import s3fs
except ImportError:
    s3fs = None


class S3Client:
    """S3 client."""

    def __init__(self, access_key, secret_key):
        """Constructor."""
        if s3fs is None:
            raise Exception("s3fs is not installed.")

        self.fs = s3fs.S3FileSystem(key=access_key, secret=secret_key)

    def read_file(self, s3_path):
        """Reads a file from S3."""
        with self.fs.open(s3_path, "rb") as f:
            return f.read()


class S3OrcidClient(S3Client):
    """S3 ORCiD client."""

    def __init__(self):
        """Constructor."""
        super().__init__(
            access_key=current_app.config["VOCABULARIES_ORCID_ACCESS_KEY"],
            secret_key=current_app.config["VOCABULARIES_ORCID_SECRET_KEY"],
        )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021-2024 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Names datastreams, transformers, writers and readers."""

import csv
import io
import tarfile
from concurrent.futures import ThreadPoolExecutor, as_completed
from datetime import timedelta

import arrow
import regex as re
from flask import current_app
from invenio_access.permissions import system_identity
from invenio_records.dictutils import dict_lookup

from invenio_vocabularies.contrib.names.s3client import S3OrcidClient

from ...datastreams.errors import TransformerError
from ...datastreams.readers import BaseReader, SimpleHTTPReader
from ...datastreams.transformers import BaseTransformer
from ...datastreams.writers import ServiceWriter


class OrcidDataSyncReader(BaseReader):
    """ORCiD Data Sync Reader."""

    def __init__(self, origin=None, mode="r", since=None, *args, **kwargs):
        """Constructor.

        :param origin: Data source (e.g. filepath).
                       Can be none in case of piped readers.
        """
        super().__init__(origin=origin, mode=mode, *args, **kwargs)
        self.s3_client = S3OrcidClient()
        self.since = since

    def _fetch_orcid_data(self, orcid_to_sync, bucket):
        """Fetches a single ORCiD record from S3."""
        # The ORCiD file key is located in a folder which name corresponds to the last three digits of the ORCiD
        suffix = orcid_to_sync[-3:]
        key = f"{suffix}/{orcid_to_sync}.xml"
        try:
            return self.s3_client.read_file(f"s3://{bucket}/{key}")
        except Exception as e:
            # TODO: log
            return None

    def _process_lambda_file(self, fileobj):
        """Process the ORCiD lambda file and returns a list of ORCiDs to sync.

        The decoded fileobj looks like the following:
        orcid, path, date_created, last_modified
        0000-0001-5109-3700, http://orcid.org/0000-0001-5109-3700, 2014-08-02 15:00:00.000,2021-08-02 15:00:00.000

        Yield ORCiDs to sync until the last sync date is reached.
        """
        date_format = "YYYY-MM-DD HH:mm:ss.SSSSSS"
        date_format_no_millis = "YYYY-MM-DD HH:mm:ss"
        time_shift = current_app.config["VOCABULARIES_ORCID_SYNC_SINCE"]
        if self.since:
            time_shift = self.since
        last_sync = arrow.now() - timedelta(**time_shift)

        file_content = fileobj.read().decode("utf-8")

        csv_reader = csv.DictReader(file_content.splitlines())

        for row in csv_reader:  # Skip the header line
            orcid = row["orcid"]

            # Lambda file is ordered by last modified date
            last_modified_str = row["last_modified"]
            try:
                last_modified_date = arrow.get(last_modified_str, date_format)
            except arrow.parser.ParserError:
                last_modified_date = arrow.get(last_modified_str, date_format_no_millis)

            if last_modified_date < last_sync:
                break
            yield orcid

    def _iter(self, orcids):
        """Iterates over the ORCiD records yielding each one."""
        with ThreadPoolExecutor(
            max_workers=current_app.config["VOCABULARIES_ORCID_SYNC_MAX_WORKERS"]
        ) as executor:
            futures = [
                executor.submit(
                    self._fetch_orcid_data,
                    orcid,
                    current_app.config["VOCABULARIES_ORCID_SUMMARIES_BUCKET"],
                )
                for orcid in orcids
            ]
            for future in as_completed(futures):
                result = future.result()
                if result is not None:
                    yield result

    def read(self, item=None, *args, **kwargs):
        """Streams the ORCiD lambda file, process it to get the ORCiDS to sync and yields it's data."""
        # Read the file from S3
        tar_content = self.s3_client.read_file(
            "s3://orcid-lambda-file/last_modified.csv.tar"
        )

        orcids_to_sync = []
        # Opens tar file and process it
        with tarfile.open(fileobj=io.BytesIO(tar_content)) as tar:
            # Iterate over each member (file or directory) in the tar file
            for member in tar.getmembers():
                # Extract the file
                extracted_file = tar.extractfile(member)
                if extracted_file:
                    # Process the file and get the ORCiDs to sync
                    orcids_to_sync.extend(self._process_lambda_file(extracted_file))

        yield from self._iter(orcids_to_sync)


class OrcidHTTPReader(SimpleHTTPReader):
    """ORCiD HTTP Reader."""

    def __init__(self, *args, test_mode=True, **kwargs):
        """Constructor."""
        if test_mode:
            origin = "https://sandbox.orcid.org/{id}"
        else:
            origin = "https://orcid.org/{id}"

        super().__init__(origin, *args, **kwargs)


DEFAULT_NAMES_EXCLUDE_REGEX = r"[\p{P}\p{S}\p{Nd}\p{No}\p{Emoji}--,.()\-']"
"""Regex to filter out names with punctuations, symbols, decimal numbers and emojis."""


class OrcidTransformer(BaseTransformer):
    """Transforms an ORCiD record into a names record."""

    def __init__(
        self, *args, names_exclude_regex=DEFAULT_NAMES_EXCLUDE_REGEX, **kwargs
    ) -> None:
        """Constructor."""
        self._names_exclude_regex = names_exclude_regex
        super().__init__()

    def _is_valid_name(self, name):
        """Check whether the name passes the regex."""
        if not self._names_exclude_regex:
            return True
        return not bool(re.search(self._names_exclude_regex, name, re.UNICODE | re.V1))

    def apply(self, stream_entry, **kwargs):
        """Applies the transformation to the stream entry."""
        record = stream_entry.entry
        person = record["person"]
        orcid_id = record["orcid-identifier"]["path"]

        name = person.get("name")
        if name is None:
            raise TransformerError(f"Name not found in ORCiD entry.")
        if name.get("family-name") is None:
            raise TransformerError(f"Family name not found in ORCiD entry.")

        if not self._is_valid_name(name["given-names"] + name["family-name"]):
            raise TransformerError(f"Invalid characters in name.")

        entry = {
            "id": orcid_id,
            "given_name": name.get("given-names"),
            "family_name": name.get("family-name"),
            "identifiers": [{"scheme": "orcid", "identifier": orcid_id}],
            "affiliations": [],
        }

        try:
            employments = dict_lookup(
                record, "activities-summary.employments.affiliation-group"
            )
            if isinstance(employments, dict):
                employments = [employments]
            history = set()
            for employment in employments:
                terminated = employment["employment-summary"].get("end-date")
                affiliation = dict_lookup(
                    employment,
                    "employment-summary.organization.name",
                )
                if affiliation not in history and not terminated:
                    history.add(affiliation)
                    entry["affiliations"].append({"name": affiliation})
        except Exception:
            pass

        stream_entry.entry = entry
        return stream_entry


class NamesServiceWriter(ServiceWriter):
    """Names service writer."""

    def __init__(self, *args, **kwargs):
        """Constructor."""
        service_or_name = kwargs.pop("service_or_name", "names")
        super().__init__(service_or_name=service_or_name, *args, **kwargs)

    def _entry_id(self, entry):
        """Get the id from an entry."""
        return entry["id"]


VOCABULARIES_DATASTREAM_READERS = {
    "orcid-http": OrcidHTTPReader,
    "orcid-data-sync": OrcidDataSyncReader,
}


VOCABULARIES_DATASTREAM_TRANSFORMERS = {
    "orcid": OrcidTransformer,
}
"""ORCiD Data Streams transformers."""


VOCABULARIES_DATASTREAM_WRITERS = {
    "names-service": NamesServiceWriter,
}
"""ORCiD Data Streams transformers."""


DATASTREAM_CONFIG = {
    "readers": [
        {
            "type": "tar",
            "args": {
                "regex": "\\.xml$",
            },
        },
        {
            "type": "xml",
            "args": {
                "root_element": "record",
            },
        },
    ],
    "transformers": [{"type": "orcid"}],
    "writers": [
        {
            "type": "names-service",
            "args": {
                "identity": system_identity,
            },
        }
    ],
}
"""ORCiD Data Stream configuration.

An origin is required for the reader.
"""

# TODO: Used on the jobs and should be set as a "PRESET" (naming to be defined)
ORCID_PRESET_DATASTREAM_CONFIG = {
    "readers": [
        {
            "type": "orcid-data-sync",
        },
        {
            "type": "xml",
            "args": {
                "root_element": "record",
            },
        },
    ],
    "transformers": [{"type": "orcid"}],
    "writers": [
        {
            "type": "async",
            "args": {
                "writer": {
                    "type": "names-service",
                }
            },
        }
    ],
    "batch_size": 1000,
    "write_many": True,
}
"""ORCiD Data Stream configuration.

An origin is required for the reader.
"""



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021-2024 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Pytest configuration.

See https://pytest-invenio.readthedocs.io/ for documentation on which test
fixtures are available.
"""

import json
import zipfile
from pathlib import Path

import pytest

from invenio_vocabularies.config import (
    VOCABULARIES_DATASTREAM_READERS,
    VOCABULARIES_DATASTREAM_TRANSFORMERS,
    VOCABULARIES_DATASTREAM_WRITERS,
)
from invenio_vocabularies.datastreams.errors import TransformerError, WriterError
from invenio_vocabularies.datastreams.readers import BaseReader, JsonReader, ZipReader
from invenio_vocabularies.datastreams.transformers import BaseTransformer
from invenio_vocabularies.datastreams.writers import BaseWriter


class TestReader(BaseReader):
    """Test reader."""

    def _iter(self, fp, *args, **kwargs):
        """Yields the values in the origin."""
        for value in self._origin:
            yield value

    def read(self, item=None, *args, **kwargs):
        """Reads from item or opens the file descriptor from origin."""
        yield from self._iter(fp=self._origin, *args, **kwargs)


class TestTransformer(BaseTransformer):
    """Test transformer."""

    def apply(self, stream_entry, *args, **kwargs):
        """Sums up one to the value."""
        if stream_entry.entry < 0:
            raise TransformerError("Value cannot be negative")

        stream_entry.entry += 1
        return stream_entry


class TestWriter(BaseWriter):
    """Test writer."""

    def write(self, stream_entry, *args, **kwargs):
        """NOP write."""
        pass

    def write_many(self, stream_entries, *args, **kwargs):
        """NOP write."""
        pass


class FailingTestWriter(BaseWriter):
    """Failing test writer."""

    def __init__(self, fail_on):
        """Initialise error."""
        super().__init__()
        self.fail_on = fail_on

    def write(self, stream_entry, *args, **kwargs):
        """Return the entry."""
        if stream_entry.entry == self.fail_on:
            raise WriterError(f"{self.fail_on} value found.")

    def write_many(self, stream_entries, *args, **kwargs):
        """NOP write."""
        pass


@pytest.fixture(scope="module")
def app_config(app_config):
    """Mimic an instance's configuration."""
    app_config["VOCABULARIES_DATASTREAM_READERS"] = {
        **VOCABULARIES_DATASTREAM_READERS,
        "test": TestReader,
    }
    app_config["VOCABULARIES_DATASTREAM_TRANSFORMERS"] = {
        **VOCABULARIES_DATASTREAM_TRANSFORMERS,
        "test": TestTransformer,
    }
    app_config["VOCABULARIES_DATASTREAM_WRITERS"] = {
        **VOCABULARIES_DATASTREAM_WRITERS,
        "test": TestWriter,
        "fail": FailingTestWriter,
    }

    return app_config


@pytest.fixture(scope="module")
def json_list():
    """Expected json list."""
    return [{"test": {"inner": "value"}}, {"test": {"inner": "value"}}]


@pytest.fixture(scope="module")
def json_element():
    """Expected json element."""
    return {"test": {"inner": "value"}}


@pytest.fixture(scope="function")
def zip_file(json_list):
    """Creates a Zip file with three files (two json) inside.

    Each iteration should return the content of one json file,
    it should ignore the .other file.
    """
    files = ["file_one.json", "file_two.json", "file_three.other"]
    filename = Path("reader_test.zip")
    with zipfile.ZipFile(file=filename, mode="w") as archive:
        for file_ in files:
            inner_filename = Path(file_)
            with open(inner_filename, "w") as file:
                json.dump(json_list, file)
            archive.write(inner_filename)
            inner_filename.unlink()

    yield filename

    filename.unlink()  # delete created file



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Affiliations Marshmallow schema tests."""

import pytest
from marshmallow import ValidationError

from invenio_vocabularies.contrib.affiliations.schema import (
    AffiliationRelationSchema,
    AffiliationSchema,
)


def test_valid_full(app, affiliation_full_data):
    loaded = AffiliationSchema().load(affiliation_full_data)
    affiliation_full_data["pid"] = affiliation_full_data.pop("id")
    assert affiliation_full_data == loaded


def test_valid_minimal(app):
    data = {
        "id": "test",
        "name": "Test affiliation",
    }
    loaded = AffiliationSchema().load(data)
    assert {
        "pid": "test",
        "name": "Test affiliation",
    } == loaded


def test_invalid_no_name(app):
    invalid = {
        "acronym": "TEST",
        "id": "aff-1",
        "identifiers": [{"identifier": "03yrm5c26", "scheme": "ror"}],
        "title": {"en": "Test affiliation", "es": "Afiliacion de test"},
    }
    with pytest.raises(ValidationError):
        data = AffiliationSchema().load(invalid)


#
# AffiliationRelationSchema
#
def test_valid_id():
    valid_id = {
        "id": "test",
    }
    assert valid_id == AffiliationRelationSchema().load(valid_id)


def test_valid_name():
    valid_name = {"name": "Entity One"}
    assert valid_name == AffiliationRelationSchema().load(valid_name)


def test_valid_both_id_name():
    valid_id_name = {"id": "test", "name": "Entity One"}
    assert valid_id_name == AffiliationRelationSchema().load(valid_id_name)


def test_invalid_empty():
    invalid_empty = {}
    with pytest.raises(ValidationError):
        data = AffiliationRelationSchema().load(invalid_empty)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2024 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Vocabulary permissions."""

from invenio_records_permissions import RecordPermissionPolicy
from invenio_records_permissions.generators import AnyUser, SystemProcess


class PermissionPolicy(RecordPermissionPolicy):
    """Permission policy."""

    can_search = [SystemProcess(), AnyUser()]
    can_read = [SystemProcess(), AnyUser()]
    can_create = [SystemProcess()]
    can_update = [SystemProcess()]
    can_delete = [SystemProcess()]
    can_manage = [SystemProcess()]
    # this permission is needed for the /api/vocabularies/ endpoint
    can_list_vocabularies = [SystemProcess(), AnyUser()]



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2024 CERN.
# Copyright (C) 2024 Uni Münster.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Vocabulary resource."""

import json

import marshmallow as ma
from flask import g
from flask_resources import (
    BaseListSchema,
    JSONSerializer,
    MarshmallowSerializer,
    ResponseHandler,
    resource_requestctx,
    response_handler,
)
from invenio_access.permissions import system_identity
from invenio_records_resources.resources import (
    RecordResource,
    RecordResourceConfig,
    SearchRequestArgsSchema,
)
from invenio_records_resources.resources.records.headers import etag_headers
from invenio_records_resources.resources.records.resource import (
    request_data,
    request_headers,
    request_search_args,
    request_view_args,
    route,
)
from invenio_records_resources.resources.records.utils import search_preference
from marshmallow import fields

from .serializer import VocabularyL10NItemSchema


#
# Resource
#
class VocabulariesResource(RecordResource):
    """Resource for generic vocabularies.

    Provide the API /api/vocabularies/
    """

    def create_url_rules(self):
        """Create the URL rules for the record resource."""
        routes = self.config.routes
        rules = super().create_url_rules()

        rules.append(
            route("POST", routes["tasks"], self.launch),
        )
        return rules

    @request_search_args
    @request_view_args
    @response_handler(many=True)
    def search(self):
        """Perform a search over the items."""
        hits = self.service.search(
            identity=g.identity,
            params=resource_requestctx.args,
            type=resource_requestctx.view_args["type"],
            search_preference=search_preference(),
        )
        return hits.to_dict(), 200

    @request_view_args
    @request_data
    @response_handler()
    def create(self):
        """Create an item."""
        item = self.service.create(
            g.identity,
            resource_requestctx.data or {},
        )
        return item.to_dict(), 201

    @request_view_args
    @response_handler()
    def read(self):
        """Read an item."""
        pid_value = (
            resource_requestctx.view_args["type"],
            resource_requestctx.view_args["pid_value"],
        )
        item = self.service.read(g.identity, pid_value)
        return item.to_dict(), 200

    @request_headers
    @request_view_args
    @request_data
    @response_handler()
    def update(self):
        """Update an item."""
        pid_value = (
            resource_requestctx.view_args["type"],
            resource_requestctx.view_args["pid_value"],
        )
        item = self.service.update(
            g.identity,
            pid_value,
            resource_requestctx.data,
            revision_id=resource_requestctx.headers.get("if_match"),
        )
        return item.to_dict(), 200

    @request_headers
    @request_view_args
    def delete(self):
        """Delete an item."""
        pid_value = (
            resource_requestctx.view_args["type"],
            resource_requestctx.view_args["pid_value"],
        )
        self.service.delete(
            g.identity,
            pid_value,
            revision_id=resource_requestctx.headers.get("if_match"),
        )
        return "", 204

    @request_data
    def launch(self):
        """Create a task."""
        self.service.launch(g.identity, resource_requestctx.data or {})
        return "", 202


class VocabulariesAdminResource(RecordResource):
    """Resource for vocabularies admin interface."""

    def create_url_rules(self):
        """Create the URL rules for the record resource."""
        routes = self.config.routes

        rules = [route("GET", routes["list"], self.search)]

        return rules

    @request_search_args
    @response_handler(many=True)
    def search(self):
        """Return information about _all_ vocabularies."""
        identity = g.identity
        hits = self.service.search(identity, params=resource_requestctx.args)

        return hits.to_dict(), 200
